# Copyright 2014-2021 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' Helpers for interacting with OpenvSwitch '''
import collections
import hashlib
import os
import re
import subprocess

from charmhelpers import deprecate
from charmhelpers.contrib.network.ovs import ovsdb as ch_ovsdb
from charmhelpers.fetch import apt_install

from charmhelpers.core.hookenv import (
    log, WARNING, INFO, DEBUG, charm_name
)
from charmhelpers.core.host import (
    CompareHostReleases,
    lsb_release,
    service
)


BRIDGE_TEMPLATE = """\
# This veth pair is required when neutron data-port is mapped to an existing linux bridge. lp:1635067

auto {linuxbridge_port}
iface {linuxbridge_port} inet manual
    pre-up ip link add name {linuxbridge_port} type veth peer name {ovsbridge_port}
    pre-up ip link set {ovsbridge_port} master {bridge}
    pre-up ip link set {ovsbridge_port} up
    up ip link set {linuxbridge_port} up
    down ip link del {linuxbridge_port}
"""

MAX_KERNEL_INTERFACE_NAME_LEN = 15


def get_bridges():
    """Return list of the bridges on the default openvswitch

    :returns: List of bridge names
    :rtype: List[str]
    :raises: subprocess.CalledProcessError if ovs-vsctl fails
    """
    cmd = ["ovs-vsctl", "list-br"]
    lines = subprocess.check_output(cmd).decode('utf-8').split("\n")
    maybe_bridges = [l.strip() for l in lines]
    return [b for b in maybe_bridges if b]


def get_bridge_ports(name):
    """Return a list the ports on a named bridge

    :param name: the name of the bridge to list
    :type name: str
    :returns: List of ports on the named bridge
    :rtype: List[str]
    :raises: subprocess.CalledProcessError if the ovs-vsctl command fails. If
        the named bridge doesn't exist, then the exception will be raised.
    """
    cmd = ["ovs-vsctl", "--", "list-ports", name]
    lines = subprocess.check_output(cmd).decode('utf-8').split("\n")
    maybe_ports = [l.strip() for l in lines]
    return [p for p in maybe_ports if p]


def get_bridges_and_ports_map():
    """Return dictionary of bridge to ports for the default openvswitch

    :returns: a mapping of bridge name to a list of ports.
    :rtype: Dict[str, List[str]]
    :raises: subprocess.CalledProcessError if any of the underlying ovs-vsctl
        command fail.
    """
    return {b: get_bridge_ports(b) for b in get_bridges()}


def _dict_to_vsctl_set(data, table, entity):
    """Helper that takes dictionary and provides ``ovs-vsctl set`` commands

    :param data: Additional data to attach to interface
        The keys in the data dictionary map directly to column names in the
        OpenvSwitch table specified as defined in DB-SCHEMA [0] referenced in
        RFC 7047 [1]

        There are some established conventions for keys in the external-ids
        column of various tables, consult the OVS Integration Guide [2] for
        more details.

        NOTE(fnordahl): Technically the ``external-ids`` column is called
        ``external_ids`` (with an underscore) and we rely on ``ovs-vsctl``'s
        behaviour of transforming dashes to underscores for us [3] so we can
        have a more pleasant data structure.

        0: http://www.openvswitch.org/ovs-vswitchd.conf.db.5.pdf
        1: https://tools.ietf.org/html/rfc7047
        2: http://docs.openvswitch.org/en/latest/topics/integration/
        3: https://github.com/openvswitch/ovs/blob/
               20dac08fdcce4b7fda1d07add3b346aa9751cfbc/
                   lib/db-ctl-base.c#L189-L215
    :type data: Optional[Dict[str,Union[str,Dict[str,str]]]]
    :param table: Name of table to operate on
    :type table: str
    :param entity: Name of entity to operate on
    :type entity: str
    :returns: '--' separated ``ovs-vsctl set`` commands
    :rtype: Iterator[Tuple[str, str, str, str, str]]
    """
    for (k, v) in data.items():
        if isinstance(v, dict):
            entries = {
                '{}:{}'.format(k, dk): dv for (dk, dv) in v.items()}
        else:
            entries = {k: v}
        for (colk, colv) in entries.items():
            yield ('--', 'set', table, entity, '{}={}'.format(colk, colv))


def add_bridge(name, datapath_type=None, brdata=None, exclusive=False):
    """Add the named bridge to openvswitch and set/update bridge data for it

    :param name: Name of bridge to create
    :type name: str
    :param datapath_type: Add datapath_type to bridge (DEPRECATED, use brdata)
    :type datapath_type: Optional[str]
    :param brdata: Additional data to attach to bridge
        The keys in the brdata dictionary map directly to column names in the
        OpenvSwitch bridge table as defined in DB-SCHEMA [0] referenced in
        RFC 7047 [1]

        There are some established conventions for keys in the external-ids
        column of various tables, consult the OVS Integration Guide [2] for
        more details.

        NOTE(fnordahl): Technically the ``external-ids`` column is called
        ``external_ids`` (with an underscore) and we rely on ``ovs-vsctl``'s
        behaviour of transforming dashes to underscores for us [3] so we can
        have a more pleasant data structure.

        0: http://www.openvswitch.org/ovs-vswitchd.conf.db.5.pdf
        1: https://tools.ietf.org/html/rfc7047
        2: http://docs.openvswitch.org/en/latest/topics/integration/
        3: https://github.com/openvswitch/ovs/blob/
               20dac08fdcce4b7fda1d07add3b346aa9751cfbc/
                   lib/db-ctl-base.c#L189-L215
    :type brdata: Optional[Dict[str,Union[str,Dict[str,str]]]]
    :param exclusive: If True, raise exception if bridge exists
    :type exclusive: bool
    :raises: subprocess.CalledProcessError
    """
    log('Creating bridge {}'.format(name))
    cmd = ['ovs-vsctl', '--']
    if not exclusive:
        cmd.append('--may-exist')
    cmd.extend(('add-br', name))
    if brdata:
        for setcmd in _dict_to_vsctl_set(brdata, 'bridge', name):
            cmd.extend(setcmd)
    if datapath_type is not None:
        log('DEPRECATION WARNING: add_bridge called with datapath_type, '
            'please use the brdata keyword argument instead.')
        cmd += ['--', 'set', 'bridge', name,
                'datapath_type={}'.format(datapath_type)]
    subprocess.check_call(cmd)


def del_bridge(name):
    """Delete the named bridge from openvswitch

    :param name: Name of bridge to remove
    :type name: str
    :raises: subprocess.CalledProcessError
    """
    log('Deleting bridge {}'.format(name))
    subprocess.check_call(["ovs-vsctl", "--", "--if-exists", "del-br", name])


def add_bridge_port(name, port, promisc=False, ifdata=None, exclusive=False,
                    linkup=True, portdata=None):
    """Add port to bridge and optionally set/update interface data for it

    :param name: Name of bridge to attach port to
    :type name: str
    :param port: Name of port as represented in netdev
    :type port: str
    :param promisc: Whether to set promiscuous mode on interface
                    True=on, False=off, None leave untouched
    :type promisc: Optional[bool]
    :param ifdata: Additional data to attach to interface
        The keys in the ifdata dictionary map directly to column names in the
        OpenvSwitch Interface table as defined in DB-SCHEMA [0] referenced in
        RFC 7047 [1]

        There are some established conventions for keys in the external-ids
        column of various tables, consult the OVS Integration Guide [2] for
        more details.

        NOTE(fnordahl): Technically the ``external-ids`` column is called
        ``external_ids`` (with an underscore) and we rely on ``ovs-vsctl``'s
        behaviour of transforming dashes to underscores for us [3] so we can
        have a more pleasant data structure.

        0: http://www.openvswitch.org/ovs-vswitchd.conf.db.5.pdf
        1: https://tools.ietf.org/html/rfc7047
        2: http://docs.openvswitch.org/en/latest/topics/integration/
        3: https://github.com/openvswitch/ovs/blob/
               20dac08fdcce4b7fda1d07add3b346aa9751cfbc/
                   lib/db-ctl-base.c#L189-L215
    :type ifdata: Optional[Dict[str,Union[str,Dict[str,str]]]]
    :param exclusive: If True, raise exception if port exists
    :type exclusive: bool
    :param linkup: Bring link up
    :type linkup: bool
    :param portdata: Additional data to attach to port. Similar to ifdata.
    :type portdata: Optional[Dict[str,Union[str,Dict[str,str]]]]
    :raises: subprocess.CalledProcessError
    """
    cmd = ['ovs-vsctl', '--']
    if not exclusive:
        cmd.append('--may-exist')
    cmd.extend(('add-port', name, port))
    for ovs_table, data in (('Interface', ifdata), ('Port', portdata)):
        if data:
            for setcmd in _dict_to_vsctl_set(data, ovs_table, port):
                cmd.extend(setcmd)

    log('Adding port {} to bridge {}'.format(port, name))
    subprocess.check_call(cmd)
    if linkup:
        # This is mostly a workaround for CI environments, in the real world
        # the bare metal provider would most likely have configured and brought
        # up the link for us.
        subprocess.check_call(["ip", "link", "set", port, "up"])
    if promisc:
        subprocess.check_call(["ip", "link", "set", port, "promisc", "on"])
    elif promisc is False:
        subprocess.check_call(["ip", "link", "set", port, "promisc", "off"])


def del_bridge_port(name, port, linkdown=True):
    """Delete a port from the named openvswitch bridge

    :param name: Name of bridge to remove port from
    :type name: str
    :param port: Name of port to remove
    :type port: str
    :param linkdown: Whether to set link down on interface (default: True)
    :type linkdown: bool
    :raises: subprocess.CalledProcessError
    """
    log('Deleting port {} from bridge {}'.format(port, name))
    subprocess.check_call(["ovs-vsctl", "--", "--if-exists", "del-port",
                           name, port])
    if linkdown:
        subprocess.check_call(["ip", "link", "set", port, "down"])
        subprocess.check_call(["ip", "link", "set", port, "promisc", "off"])


def add_bridge_bond(bridge, port, interfaces, portdata=None, ifdatamap=None,
                    exclusive=False):
    """Add bonded port in bridge from interfaces.

    :param bridge: Name of bridge to add bonded port to
    :type bridge: str
    :param port: Name of created port
    :type port: str
    :param interfaces: Underlying interfaces that make up the bonded port
    :type interfaces: Iterator[str]
    :param portdata: Additional data to attach to the created bond port
        See _dict_to_vsctl_set() for detailed description.
        Example:
        {
            'bond-mode': 'balance-tcp',
            'lacp': 'active',
            'other-config': {
                'lacp-time': 'fast',
            },
        }
    :type portdata: Optional[Dict[str,Union[str,Dict[str,str]]]]
    :param ifdatamap: Map of data to attach to created bond interfaces
        See _dict_to_vsctl_set() for detailed description.
        Example:
        {
            'eth0': {
                'type': 'dpdk',
                'mtu-request': '9000',
                'options': {
                    'dpdk-devargs': '0000:01:00.0',
                },
            },
        }
    :type ifdatamap: Optional[Dict[str,Dict[str,Union[str,Dict[str,str]]]]]
    :param exclusive: If True, raise exception if port exists
    :type exclusive: bool
    :raises: subprocess.CalledProcessError
    """
    cmd = ['ovs-vsctl', '--']
    if not exclusive:
        cmd.append('--may-exist')
    cmd.extend(('add-bond', bridge, port))
    cmd.extend(interfaces)
    if portdata:
        for setcmd in _dict_to_vsctl_set(portdata, 'port', port):
            cmd.extend(setcmd)
    if ifdatamap:
        for ifname, ifdata in ifdatamap.items():
            for setcmd in _dict_to_vsctl_set(ifdata, 'Interface', ifname):
                cmd.extend(setcmd)
    subprocess.check_call(cmd)


@deprecate('see lp:1877594', '2021-01', log=log)
def add_ovsbridge_linuxbridge(name, bridge, ifdata=None, portdata=None):
    """Add linux bridge to the named openvswitch bridge

    :param name: Name of ovs bridge to be added to Linux bridge
    :type name: str
    :param bridge: Name of Linux bridge to be added to ovs bridge
    :type name: str
    :param ifdata: Additional data to attach to interface
        The keys in the ifdata dictionary map directly to column names in the
        OpenvSwitch Interface table as defined in DB-SCHEMA [0] referenced in
        RFC 7047 [1]

        There are some established conventions for keys in the external-ids
        column of various tables, consult the OVS Integration Guide [2] for
        more details.

        NOTE(fnordahl): Technically the ``external-ids`` column is called
        ``external_ids`` (with an underscore) and we rely on ``ovs-vsctl``'s
        behaviour of transforming dashes to underscores for us [3] so we can
        have a more pleasant data structure.

        0: http://www.openvswitch.org/ovs-vswitchd.conf.db.5.pdf
        1: https://tools.ietf.org/html/rfc7047
        2: http://docs.openvswitch.org/en/latest/topics/integration/
        3: https://github.com/openvswitch/ovs/blob/
               20dac08fdcce4b7fda1d07add3b346aa9751cfbc/
                   lib/db-ctl-base.c#L189-L215
    :type ifdata: Optional[Dict[str,Union[str,Dict[str,str]]]]
    :param portdata: Additional data to attach to port. Similar to ifdata.
    :type portdata: Optional[Dict[str,Union[str,Dict[str,str]]]]

    WARNINGS:
    * The `ifup` command (NetworkManager) must be available on the system for
      this to work. Before bionic this was shipped by default. On bionic and
      newer you need to install the package `ifupdown`. This might however cause
      issues when deploying to LXD, see lp:1877594, which is why this function
      isn't supported anymore.
    * On focal and newer this function won't even try to run `ifup` and raise
      directly.
    """
    try:
        import netifaces
    except ImportError:
        apt_install('python3-netifaces', fatal=True)
        import netifaces

    # NOTE(jamespage):
    # Older code supported addition of a linuxbridge directly
    # to an OVS bridge; ensure we don't break uses on upgrade
    existing_ovs_bridge = port_to_br(bridge)
    if existing_ovs_bridge is not None:
        log('Linuxbridge {} is already directly in use'
            ' by OVS bridge {}'.format(bridge, existing_ovs_bridge),
            level=INFO)
        return

    # NOTE(jamespage):
    # preserve existing naming because interfaces may already exist.
    ovsbridge_port = "veth-" + name
    linuxbridge_port = "veth-" + bridge
    if (len(ovsbridge_port) > MAX_KERNEL_INTERFACE_NAME_LEN or
            len(linuxbridge_port) > MAX_KERNEL_INTERFACE_NAME_LEN):
        # NOTE(jamespage):
        # use parts of hashed bridgename (openstack style) when
        # a bridge name exceeds 15 chars
        hashed_bridge = hashlib.sha256(bridge.encode('UTF-8')).hexdigest()
        base = '{}-{}'.format(hashed_bridge[:8], hashed_bridge[-2:])
        ovsbridge_port = "cvo{}".format(base)
        linuxbridge_port = "cvb{}".format(base)

    network_interface_already_exists = False
    interfaces = netifaces.interfaces()
    for interface in interfaces:
        if interface == ovsbridge_port or interface == linuxbridge_port:
            log('Interface {} already exists'.format(interface), level=INFO)
            network_interface_already_exists = True
            break

    log('Adding linuxbridge {} to ovsbridge {}'.format(bridge, name),
        level=INFO)

    if not network_interface_already_exists:
        setup_eni()  # will raise on focal+

        with open('/etc/network/interfaces.d/{}.cfg'.format(
                linuxbridge_port), 'w') as config:
            config.write(BRIDGE_TEMPLATE.format(
                linuxbridge_port=linuxbridge_port,
                ovsbridge_port=ovsbridge_port, bridge=bridge))

        try:
            # NOTE(lourot): 'ifup <name>' can't be replaced by
            # 'ip link set <name> up' as the latter won't parse
            # /etc/network/interfaces*
            subprocess.check_call(['ifup', linuxbridge_port])
        except FileNotFoundError:
            # NOTE(lourot): on bionic and newer, 'ifup' isn't installed by
            # default. It has been replaced by netplan.io but we can't use it
            # yet because of lp:1876730. For the time being, charms using this
            # have to install 'ifupdown' on bionic and newer. This will however
            # cause issues when deploying to LXD, see lp:1877594.
            raise RuntimeError('ifup: command not found. Did this charm forget '
                               'to install ifupdown?')

    add_bridge_port(name, linuxbridge_port, ifdata=ifdata, exclusive=False,
                    portdata=portdata)


def is_linuxbridge_interface(port):
    ''' Check if the interface is a linuxbridge bridge
    :param port: Name of an interface to check whether it is a Linux bridge
    :returns: True if port is a Linux bridge'''

    if os.path.exists('/sys/class/net/' + port + '/bridge'):
        log('Interface {} is a Linux bridge'.format(port), level=DEBUG)
        return True
    else:
        log('Interface {} is not a Linux bridge'.format(port), level=DEBUG)
        return False


def set_manager(manager):
    ''' Set the controller for the local openvswitch '''
    log('Setting manager for local ovs to {}'.format(manager))
    subprocess.check_call(['ovs-vsctl', 'set-manager',
                           'ssl:{}'.format(manager)])


def set_Open_vSwitch_column_value(column_value):
    """
    Calls ovs-vsctl and sets the 'column_value' in the Open_vSwitch table.

    :param column_value:
            See http://www.openvswitch.org//ovs-vswitchd.conf.db.5.pdf for
            details of the relevant values.
    :type str
    :raises CalledProcessException: possibly ovsdb-server is not running
    """
    log('Setting {} in the Open_vSwitch table'.format(column_value))
    subprocess.check_call(['ovs-vsctl', 'set', 'Open_vSwitch', '.', column_value])


CERT_PATH = '/etc/openvswitch/ovsclient-cert.pem'


def get_certificate():
    ''' Read openvswitch certificate from disk '''
    if os.path.exists(CERT_PATH):
        log('Reading ovs certificate from {}'.format(CERT_PATH))
        with open(CERT_PATH, 'r') as cert:
            full_cert = cert.read()
            begin_marker = "-----BEGIN CERTIFICATE-----"
            end_marker = "-----END CERTIFICATE-----"
            begin_index = full_cert.find(begin_marker)
            end_index = full_cert.rfind(end_marker)
            if end_index == -1 or begin_index == -1:
                raise RuntimeError("Certificate does not contain valid begin"
                                   " and end markers.")
            full_cert = full_cert[begin_index:(end_index + len(end_marker))]
            return full_cert
    else:
        log('Certificate not found', level=WARNING)
        return None


@deprecate('see lp:1877594', '2021-01', log=log)
def setup_eni():
    """Makes sure /etc/network/interfaces.d/ exists and will be parsed.

    When setting up interfaces, Juju removes from
    /etc/network/interfaces the line sourcing interfaces.d/

    WARNING: Not supported on focal and newer anymore. Will raise.
    """
    release = CompareHostReleases(lsb_release()['DISTRIB_CODENAME'])
    if release >= 'focal':
        raise RuntimeError("NetworkManager isn't supported anymore")

    if not os.path.exists('/etc/network/interfaces.d'):
        os.makedirs('/etc/network/interfaces.d', mode=0o755)
    with open('/etc/network/interfaces', 'r') as eni:
        for line in eni:
            if re.search(r'^\s*source\s+/etc/network/interfaces.d/\*\s*$',
                         line):
                return
    with open('/etc/network/interfaces', 'a') as eni:
        eni.write('\nsource /etc/network/interfaces.d/*')


@deprecate('use setup_eni() instead', '2021-01', log=log)
def check_for_eni_source():
    setup_eni()


def full_restart():
    ''' Full restart and reload of openvswitch '''
    if os.path.exists('/etc/init/openvswitch-force-reload-kmod.conf'):
        service('start', 'openvswitch-force-reload-kmod')
    else:
        service('force-reload-kmod', 'openvswitch-switch')


def enable_ipfix(bridge, target,
                 cache_active_timeout=60,
                 cache_max_flows=128,
                 sampling=64):
    '''Enable IPFIX on bridge to target.
    :param bridge: Bridge to monitor
    :param target: IPFIX remote endpoint
    :param cache_active_timeout: The maximum period in seconds for
                                 which an IPFIX flow record is cached
                                 and aggregated before being sent
    :param cache_max_flows: The maximum number of IPFIX flow records
                            that can be cached at a time
    :param sampling: The rate at which packets should be sampled and
                     sent to each target collector
    '''
    cmd = [
        'ovs-vsctl', 'set', 'Bridge', bridge, 'ipfix=@i', '--',
        '--id=@i', 'create', 'IPFIX',
        'targets="{}"'.format(target),
        'sampling={}'.format(sampling),
        'cache_active_timeout={}'.format(cache_active_timeout),
        'cache_max_flows={}'.format(cache_max_flows),
    ]
    log('Enabling IPfix on {}.'.format(bridge))
    subprocess.check_call(cmd)


def disable_ipfix(bridge):
    '''Diable IPFIX on target bridge.
    :param bridge: Bridge to modify
    '''
    cmd = ['ovs-vsctl', 'clear', 'Bridge', bridge, 'ipfix']
    subprocess.check_call(cmd)


def port_to_br(port):
    '''Determine the bridge that contains a port
    :param port: Name of port to check for
    :returns str: OVS bridge containing port or None if not found
    '''
    try:
        return subprocess.check_output(
            ['ovs-vsctl', 'port-to-br', port]
        ).decode('UTF-8').strip()
    except subprocess.CalledProcessError:
        return None


def ovs_appctl(target, args):
    """Run `ovs-appctl` for target with args and return output.

    :param target: Name of daemon to contact.  Unless target begins with '/',
                   `ovs-appctl` looks for a pidfile and will build the path to
                   a /var/run/openvswitch/target.pid.ctl for you.
    :type target: str
    :param args: Command and arguments to pass to `ovs-appctl`
    :type args: Tuple[str, ...]
    :returns: Output from command
    :rtype: str
    :raises: subprocess.CalledProcessError
    """
    cmd = ['ovs-appctl', '-t', target]
    cmd.extend(args)
    return subprocess.check_output(cmd, universal_newlines=True)


def uuid_for_port(port_name):
    """Get UUID of named port.

    :param port_name: Name of port.
    :type port_name: str
    :returns: Port UUID.
    :rtype: Optional[uuid.UUID]
    """
    for port in ch_ovsdb.SimpleOVSDB(
            'ovs-vsctl').port.find('name={}'.format(port_name)):
        return port['_uuid']


def bridge_for_port(port_uuid):
    """Find which bridge a port is on.

    :param port_uuid: UUID of port.
    :type port_uuid: uuid.UUID
    :returns: Name of bridge or None.
    :rtype: Optional[str]
    """
    for bridge in ch_ovsdb.SimpleOVSDB(
            'ovs-vsctl').bridge:
        # If there is a single port on a bridge the ports property will not be
        # a list. ref: juju/charm-helpers#510
        if (isinstance(bridge['ports'], list) and
                port_uuid in bridge['ports'] or
                port_uuid == bridge['ports']):
            return bridge['name']


PatchPort = collections.namedtuple('PatchPort', ('bridge', 'port'))
Patch = collections.namedtuple('Patch', ('this_end', 'other_end'))


def patch_ports_on_bridge(bridge):
    """Find patch ports on a bridge.

    :param bridge: Name of bridge
    :type bridge: str
    :returns: Iterator with bridge and port name for both ends of a patch.
    :rtype: Iterator[Patch[PatchPort[str,str],PatchPort[str,str]]]
    :raises: ValueError
    """
    # On any given vSwitch there will be a small number of patch ports, so we
    # start by iterating over ports with type `patch` then look up which bridge
    # they belong to and act on any ports that match the criteria.
    for interface in ch_ovsdb.SimpleOVSDB(
            'ovs-vsctl').interface.find('type=patch'):
        for port in ch_ovsdb.SimpleOVSDB(
                'ovs-vsctl').port.find('name={}'.format(interface['name'])):
            if bridge_for_port(port['_uuid']) == bridge:
                this_end = PatchPort(bridge, port['name'])
                other_end = PatchPort(bridge_for_port(
                    uuid_for_port(
                        interface['options']['peer'])),
                    interface['options']['peer'])
                yield(Patch(this_end, other_end))
            # We expect one result and it is ok if it turns out to be a port
            # for a different bridge. However we need a break here to satisfy
            # the for/else check which is in place to detect interface referring
            # to non-existent port.
            break
        else:
            raise ValueError('Port for interface named "{}" does unexpectedly '
                             'not exist.'.format(interface['name']))
    else:
        # Allow our caller to handle no patch ports found gracefully, in
        # reference to PEP479 just doing a return will provide a empty iterator
        # and not None.
        return


def generate_external_ids(external_id_value=None):
    """Generate external-ids dictionary that can be used to mark OVS bridges
    and ports as managed by the charm.

    :param external_id_value: Value of the external-ids entry.
                              Note: 'managed' will be used if not specified.
    :type external_id_value: Optional[str]
    :returns: Dict with a single external-ids entry.
              {
                  'external-ids': {
                      charm-``charm_name``: ``external_id_value``
                  }
              }
    :rtype: Dict[str, Dict[str]]
    """
    external_id_key = "charm-{}".format(charm_name())
    external_id_value = ('managed' if external_id_value is None
                         else external_id_value)
    return {
        'external-ids': {
            external_id_key: external_id_value
        }
    }



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import subprocess

from charmhelpers.core.hookenv import (
    log,
    INFO,
    ERROR,
)

from charmhelpers.contrib.hardening.audits import BaseAudit


class DisabledModuleAudit(BaseAudit):
    """Audits Apache2 modules.

    Determines if the apache2 modules are enabled. If the modules are enabled
    then they are removed in the ensure_compliance.
    """
    def __init__(self, modules):
        if modules is None:
            self.modules = []
        elif isinstance(modules, str):
            self.modules = [modules]
        else:
            self.modules = modules

    def ensure_compliance(self):
        """Ensures that the modules are not loaded."""
        if not self.modules:
            return

        try:
            loaded_modules = self._get_loaded_modules()
            non_compliant_modules = []
            for module in self.modules:
                if module in loaded_modules:
                    log("Module '%s' is enabled but should not be." %
                        (module), level=INFO)
                    non_compliant_modules.append(module)

            if len(non_compliant_modules) == 0:
                return

            for module in non_compliant_modules:
                self._disable_module(module)
            self._restart_apache()
        except subprocess.CalledProcessError as e:
            log('Error occurred auditing apache module compliance. '
                'This may have been already reported. '
                'Output is: %s' % e.output, level=ERROR)

    @staticmethod
    def _get_loaded_modules():
        """Returns the modules which are enabled in Apache."""
        output = subprocess.check_output(['apache2ctl', '-M']).decode('utf-8')
        modules = []
        for line in output.splitlines():
            # Each line of the enabled module output looks like:
            #  module_name (static|shared)
            # Plus a header line at the top of the output which is stripped
            # out by the regex.
            matcher = re.search(r'^ (\S*)_module (\S*)', line)
            if matcher:
                modules.append(matcher.group(1))
        return modules

    @staticmethod
    def _disable_module(module):
        """Disables the specified module in Apache."""
        try:
            subprocess.check_call(['a2dismod', module])
        except subprocess.CalledProcessError as e:
            # Note: catch error here to allow the attempt of disabling
            # multiple modules in one go rather than failing after the
            # first module fails.
            log('Error occurred disabling module %s. '
                'Output is: %s' % (module, e.output), level=ERROR)

    @staticmethod
    def _restart_apache():
        """Restarts the apache process"""
        subprocess.check_output(['service', 'apache2', 'restart'])

    @staticmethod
    def is_ssl_enabled():
        """Check if SSL module is enabled or not"""
        return 'ssl' in DisabledModuleAudit._get_loaded_modules()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

TRUTHY_STRINGS = {'y', 'yes', 'true', 't', 'on'}
FALSEY_STRINGS = {'n', 'no', 'false', 'f', 'off'}


def bool_from_string(value, truthy_strings=TRUTHY_STRINGS, falsey_strings=FALSEY_STRINGS, assume_false=False):
    """Interpret string value as boolean.

    Returns True if value translates to True otherwise False.
    """
    if isinstance(value, str):
        value = str(value)
    else:
        msg = "Unable to interpret non-string value '%s' as boolean" % (value)
        raise ValueError(msg)

    value = value.strip().lower()

    if value in truthy_strings:
        return True
    elif value in falsey_strings or assume_false:
        return False

    msg = "Unable to interpret string value '%s' as boolean" % (value)
    raise ValueError(msg)


def bytes_from_string(value):
    """Interpret human readable string value as bytes.

    Returns int
    """
    BYTE_POWER = {
        'K': 1,
        'KB': 1,
        'M': 2,
        'MB': 2,
        'G': 3,
        'GB': 3,
        'T': 4,
        'TB': 4,
        'P': 5,
        'PB': 5,
    }
    if isinstance(value, str):
        value = str(value)
    else:
        msg = "Unable to interpret non-string value '%s' as bytes" % (value)
        raise ValueError(msg)
    matches = re.match("([0-9]+)([a-zA-Z]+)", value)
    if matches:
        size = int(matches.group(1)) * (1024 ** BYTE_POWER[matches.group(2)])
    else:
        # Assume that value passed in is bytes
        try:
            size = int(value)
        except ValueError:
            msg = "Unable to interpret string value '%s' as bytes" % (value)
            raise ValueError(msg)
    return size


class BasicStringComparator(object):
    """Provides a class that will compare strings from an iterator type object.
    Used to provide > and < comparisons on strings that may not necessarily be
    alphanumerically ordered.  e.g. OpenStack or Ubuntu releases AFTER the
    z-wrap.
    """

    _list = None

    def __init__(self, item):
        if self._list is None:
            raise Exception("Must define the _list in the class definition!")
        try:
            self.index = self._list.index(item)
        except Exception:
            raise KeyError("Item '{}' is not in list '{}'"
                           .format(item, self._list))

    def __eq__(self, other):
        assert isinstance(other, str) or isinstance(other, self.__class__)
        return self.index == self._list.index(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        assert isinstance(other, str) or isinstance(other, self.__class__)
        return self.index < self._list.index(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def __gt__(self, other):
        assert isinstance(other, str) or isinstance(other, self.__class__)
        return self.index > self._list.index(other)

    def __le__(self, other):
        return not self.__gt__(other)

    def __str__(self):
        """Always give back the item at the index so it can be used in
        comparisons like:

        s_mitaka = CompareOpenStack('mitaka')
        s_newton = CompareOpenstack('newton')

        assert s_newton > s_mitaka

        @returns: <string>
        """
        return self._list[self.index]



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
warnings.warn("contrib.charmhelpers is deprecated", DeprecationWarning)  # noqa

import operator
import tempfile
import time
import yaml
import subprocess

from urllib.request import urlopen
from urllib.error import (HTTPError, URLError)


"""Helper functions for writing Juju charms in Python."""

__metaclass__ = type
__all__ = [
    # 'get_config',             # core.hookenv.config()
    # 'log',                    # core.hookenv.log()
    # 'log_entry',              # core.hookenv.log()
    # 'log_exit',               # core.hookenv.log()
    # 'relation_get',           # core.hookenv.relation_get()
    # 'relation_set',           # core.hookenv.relation_set()
    # 'relation_ids',           # core.hookenv.relation_ids()
    # 'relation_list',          # core.hookenv.relation_units()
    # 'config_get',             # core.hookenv.config()
    # 'unit_get',               # core.hookenv.unit_get()
    # 'open_port',              # core.hookenv.open_port()
    # 'close_port',             # core.hookenv.close_port()
    # 'service_control',        # core.host.service()
    'unit_info',              # client-side, NOT IMPLEMENTED
    'wait_for_machine',       # client-side, NOT IMPLEMENTED
    'wait_for_page_contents',  # client-side, NOT IMPLEMENTED
    'wait_for_relation',      # client-side, NOT IMPLEMENTED
    'wait_for_unit',          # client-side, NOT IMPLEMENTED
]


SLEEP_AMOUNT = 0.1


# We create a juju_status Command here because it makes testing much,
# much easier.
def juju_status():
    subprocess.check_call(['juju', 'status'])

# re-implemented as charmhelpers.fetch.configure_sources()
# def configure_source(update=False):
#    source = config_get('source')
#    if ((source.startswith('ppa:') or
#         source.startswith('cloud:') or
#         source.startswith('http:'))):
#        run('add-apt-repository', source)
#    if source.startswith("http:"):
#        run('apt-key', 'import', config_get('key'))
#    if update:
#        run('apt-get', 'update')


# DEPRECATED: client-side only
def make_charm_config_file(charm_config):
    charm_config_file = tempfile.NamedTemporaryFile(mode='w+')
    charm_config_file.write(yaml.dump(charm_config))
    charm_config_file.flush()
    # The NamedTemporaryFile instance is returned instead of just the name
    # because we want to take advantage of garbage collection-triggered
    # deletion of the temp file when it goes out of scope in the caller.
    return charm_config_file


# DEPRECATED: client-side only
def unit_info(service_name, item_name, data=None, unit=None):
    if data is None:
        data = yaml.safe_load(juju_status())
    service = data['services'].get(service_name)
    if service is None:
        # XXX 2012-02-08 gmb:
        #     This allows us to cope with the race condition that we
        #     have between deploying a service and having it come up in
        #     `juju status`. We could probably do with cleaning it up so
        #     that it fails a bit more noisily after a while.
        return ''
    units = service['units']
    if unit is not None:
        item = units[unit][item_name]
    else:
        # It might seem odd to sort the units here, but we do it to
        # ensure that when no unit is specified, the first unit for the
        # service (or at least the one with the lowest number) is the
        # one whose data gets returned.
        sorted_unit_names = sorted(units.keys())
        item = units[sorted_unit_names[0]][item_name]
    return item


# DEPRECATED: client-side only
def get_machine_data():
    return yaml.safe_load(juju_status())['machines']


# DEPRECATED: client-side only
def wait_for_machine(num_machines=1, timeout=300):
    """Wait `timeout` seconds for `num_machines` machines to come up.

    This wait_for... function can be called by other wait_for functions
    whose timeouts might be too short in situations where only a bare
    Juju setup has been bootstrapped.

    :return: A tuple of (num_machines, time_taken). This is used for
             testing.
    """
    # You may think this is a hack, and you'd be right. The easiest way
    # to tell what environment we're working in (LXC vs EC2) is to check
    # the dns-name of the first machine. If it's localhost we're in LXC
    # and we can just return here.
    if get_machine_data()[0]['dns-name'] == 'localhost':
        return 1, 0
    start_time = time.time()
    while True:
        # Drop the first machine, since it's the Zookeeper and that's
        # not a machine that we need to wait for. This will only work
        # for EC2 environments, which is why we return early above if
        # we're in LXC.
        machine_data = get_machine_data()
        non_zookeeper_machines = [
            machine_data[key] for key in list(machine_data.keys())[1:]]
        if len(non_zookeeper_machines) >= num_machines:
            all_machines_running = True
            for machine in non_zookeeper_machines:
                if machine.get('instance-state') != 'running':
                    all_machines_running = False
                    break
            if all_machines_running:
                break
        if time.time() - start_time >= timeout:
            raise RuntimeError('timeout waiting for service to start')
        time.sleep(SLEEP_AMOUNT)
    return num_machines, time.time() - start_time


# DEPRECATED: client-side only
def wait_for_unit(service_name, timeout=480):
    """Wait `timeout` seconds for a given service name to come up."""
    wait_for_machine(num_machines=1)
    start_time = time.time()
    while True:
        state = unit_info(service_name, 'agent-state')
        if 'error' in state or state == 'started':
            break
        if time.time() - start_time >= timeout:
            raise RuntimeError('timeout waiting for service to start')
        time.sleep(SLEEP_AMOUNT)
    if state != 'started':
        raise RuntimeError('unit did not start, agent-state: ' + state)


# DEPRECATED: client-side only
def wait_for_relation(service_name, relation_name, timeout=120):
    """Wait `timeout` seconds for a given relation to come up."""
    start_time = time.time()
    while True:
        relation = unit_info(service_name, 'relations').get(relation_name)
        if relation is not None and relation['state'] == 'up':
            break
        if time.time() - start_time >= timeout:
            raise RuntimeError('timeout waiting for relation to be up')
        time.sleep(SLEEP_AMOUNT)


# DEPRECATED: client-side only
def wait_for_page_contents(url, contents, timeout=120, validate=None):
    if validate is None:
        validate = operator.contains
    start_time = time.time()
    while True:
        try:
            stream = urlopen(url)
        except (HTTPError, URLError):
            pass
        else:
            page = stream.read()
            if validate(page, contents):
                return page
        if time.time() - start_time >= timeout:
            raise RuntimeError('timeout waiting for contents of ' + url)
        time.sleep(SLEEP_AMOUNT)



import mock
import textwrap
import uuid

import charmhelpers.contrib.network.ovs.ovsdb as ovsdb

import tests.utils as test_utils


VSCTL_BRIDGE_TBL = textwrap.dedent("""
    {"data":[[["uuid","1e21ba48-61ff-4b32-b35e-cb80411da351"],
    ["set",[]],["set",[]],"0000a0369fdd3890","","<unknown>",
    ["map",[["charm-ovn-chassis","managed"],["other","value"]]],
    ["set",[]],["set",[]],["map",[]],["set",[]],false,["set",[]],
    "br-test",["set",[]],["map",[]],["set",
    [["uuid","617f9359-77e2-41be-8af6-4c44e7a6bcc3"],
    ["uuid","da840476-8809-4107-8733-591f4696f056"]]],
    ["set",["OpenFlow10","OpenFlow13","OpenFlow14"]],false,["map",[]],
    ["set",[]],["map",[]],false],
    [["uuid","bb685b0f-a383-40a1-b7a5-b5c2066bfa42"],
    ["set",[]],["set",[]],"00000e5b68bba140","","<unknown>",
    ["map",[]],"secure",["set",[]],["map",[]],["set",[]],false,
    ["set",[]],"br-int",["set",[]],["map",[["disable-in-band","true"]]],
    ["set",[["uuid","07f4c231-9fd2-49b0-a558-5b69d657fdb0"],
    ["uuid","8bbd2441-866f-4317-a284-09491702776c"],
    ["uuid","d9e9c081-6482-4006-b7d6-239182b56c2e"]]],
    ["set",[]],false,["map",[]],["set",[]],["map",[]],false]],
    "headings":["_uuid","auto_attach","controller","datapath_id",
    "datapath_type","datapath_version","external_ids","fail_mode",
    "flood_vlans","flow_tables","ipfix","mcast_snooping_enable",
    "mirrors","name","netflow","other_config","ports","protocols",
    "rstp_enable","rstp_status","sflow","status","stp_enable"]}
    """)


class TestSimpleOVSDB(test_utils.BaseTestCase):

    def patch_target(self, attr, return_value=None):
        mocked = mock.patch.object(self.target, attr)
        self._patches[attr] = mocked
        started = mocked.start()
        started.return_value = return_value
        self._patches_start[attr] = started
        setattr(self, attr, started)

    def test___init__(self):
        with self.assertRaises(RuntimeError):
            self.target = ovsdb.SimpleOVSDB('atool')
        with self.assertRaises(AttributeError):
            self.target = ovsdb.SimpleOVSDB('ovs-vsctl')
            self.target.unknown_table.find()

    def test__find_tbl(self):
        self.target = ovsdb.SimpleOVSDB('ovs-vsctl')
        self.patch_object(ovsdb.utils, '_run')
        self._run.return_value = VSCTL_BRIDGE_TBL
        self.maxDiff = None
        expect = {
            '_uuid': uuid.UUID('1e21ba48-61ff-4b32-b35e-cb80411da351'),
            'auto_attach': [],
            'controller': [],
            'datapath_id': '0000a0369fdd3890',
            'datapath_type': '',
            'datapath_version': '<unknown>',
            'external_ids': {
                'charm-ovn-chassis': 'managed',
                'other': 'value',
            },
            'fail_mode': [],
            'flood_vlans': [],
            'flow_tables': {},
            'ipfix': [],
            'mcast_snooping_enable': False,
            'mirrors': [],
            'name': 'br-test',
            'netflow': [],
            'other_config': {},
            'ports': [uuid.UUID('617f9359-77e2-41be-8af6-4c44e7a6bcc3'),
                      uuid.UUID('da840476-8809-4107-8733-591f4696f056')],
            'protocols': ['OpenFlow10', 'OpenFlow13', 'OpenFlow14'],
            'rstp_enable': False,
            'rstp_status': {},
            'sflow': [],
            'status': {},
            'stp_enable': False}
        # this in effect also tests the __iter__ front end method
        for el in self.target.bridge:
            self.assertDictEqual(el, expect)
            break
        self._run.assert_called_once_with(
            'ovs-vsctl', '-f', 'json', 'find', 'bridge')
        self._run.reset_mock()
        # this in effect also tests the find front end method
        for el in self.target.bridge.find(condition='name=br-test'):
            break
        self._run.assert_called_once_with(
            'ovs-vsctl', '-f', 'json', 'find', 'bridge', 'name=br-test')
        # check the optional args paramenter
        self._run.reset_mock()
        self.target = ovsdb.SimpleOVSDB('ovs-vsctl', args=['extra', 'args'])
        for el in self.target.bridge.find(condition='name=br-test'):
            break
        self._run.assert_called_once_with(
            'ovs-vsctl', 'extra', 'args',
            '-f', 'json', 'find', 'bridge', 'name=br-test')

    def test_clear(self):
        self.target = ovsdb.SimpleOVSDB('ovs-vsctl')
        self.patch_object(ovsdb.utils, '_run')
        self.target.interface.clear('1e21ba48-61ff-4b32-b35e-cb80411da351',
                                    'external_ids')
        self._run.assert_called_once_with(
            'ovs-vsctl', 'clear', 'interface',
            '1e21ba48-61ff-4b32-b35e-cb80411da351', 'external_ids')

    def test_remove(self):
        self.target = ovsdb.SimpleOVSDB('ovs-vsctl')
        self.patch_object(ovsdb.utils, '_run')
        self.target.interface.remove('1e21ba48-61ff-4b32-b35e-cb80411da351',
                                     'external_ids', 'other')
        self._run.assert_called_once_with(
            'ovs-vsctl', 'remove', 'interface',
            '1e21ba48-61ff-4b32-b35e-cb80411da351', 'external_ids', 'other')

    def test_set(self):
        self.target = ovsdb.SimpleOVSDB('ovs-vsctl')
        self.patch_object(ovsdb.utils, '_run')
        self.target.interface.set('1e21ba48-61ff-4b32-b35e-cb80411da351',
                                  'external_ids:other', 'value')
        self._run.assert_called_once_with(
            'ovs-vsctl', 'set', 'interface',
            '1e21ba48-61ff-4b32-b35e-cb80411da351', 'external_ids:other=value')



#!/usr/bin/python

# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Authors:
#   Adam Gandelman <adamg@ubuntu.com>

import logging
import optparse
import os
import subprocess
import shutil
import sys
import tempfile
import yaml
from fnmatch import fnmatch

CHARM_HELPERS_REPO = 'https://github.com/juju/charm-helpers'


def parse_config(conf_file):
    if not os.path.isfile(conf_file):
        logging.error('Invalid config file: %s.' % conf_file)
        return False
    return yaml.load(open(conf_file).read())


def clone_helpers(work_dir, repo):
    dest = os.path.join(work_dir, 'charm-helpers')
    logging.info('Cloning out %s to %s.' % (repo, dest))
    branch = None
    if '@' in repo:
        repo, branch = repo.split('@', 1)
    cmd = ['git', 'clone', '--depth=1']
    if branch is not None:
        cmd += ['--branch', branch]
    cmd += [repo, dest]
    subprocess.check_call(cmd)
    return dest


def _module_path(module):
    return os.path.join(*module.split('.'))


def _src_path(src, module):
    return os.path.join(src, 'charmhelpers', _module_path(module))


def _dest_path(dest, module):
    return os.path.join(dest, _module_path(module))


def _is_pyfile(path):
    return os.path.isfile(path + '.py')


def ensure_init(path):
    '''
    ensure directories leading up to path are importable, omitting
    parent directory, eg path='/hooks/helpers/foo'/:
        hooks/
        hooks/helpers/__init__.py
        hooks/helpers/foo/__init__.py
    '''
    for d, dirs, files in os.walk(os.path.join(*path.split('/')[:2])):
        _i = os.path.join(d, '__init__.py')
        if not os.path.exists(_i):
            logging.info('Adding missing __init__.py: %s' % _i)
            open(_i, 'wb').close()


def sync_pyfile(src, dest):
    src = src + '.py'
    src_dir = os.path.dirname(src)
    logging.info('Syncing pyfile: %s -> %s.' % (src, dest))
    if not os.path.exists(dest):
        os.makedirs(dest)
    shutil.copy(src, dest)
    if os.path.isfile(os.path.join(src_dir, '__init__.py')):
        shutil.copy(os.path.join(src_dir, '__init__.py'),
                    dest)
    ensure_init(dest)


def get_filter(opts=None):
    opts = opts or []
    if 'inc=*' in opts:
        # do not filter any files, include everything
        return None

    def _filter(dir, ls):
        incs = [opt.split('=').pop() for opt in opts if 'inc=' in opt]
        _filter = []
        for f in ls:
            _f = os.path.join(dir, f)

            if not os.path.isdir(_f) and not _f.endswith('.py') and incs:
                if True not in [fnmatch(_f, inc) for inc in incs]:
                    logging.debug('Not syncing %s, does not match include '
                                  'filters (%s)' % (_f, incs))
                    _filter.append(f)
                else:
                    logging.debug('Including file, which matches include '
                                  'filters (%s): %s' % (incs, _f))
            elif (os.path.isfile(_f) and not _f.endswith('.py')):
                logging.debug('Not syncing file: %s' % f)
                _filter.append(f)
            elif (os.path.isdir(_f) and not
                  os.path.isfile(os.path.join(_f, '__init__.py'))):
                logging.debug('Not syncing directory: %s' % f)
                _filter.append(f)
        return _filter
    return _filter


def sync_directory(src, dest, opts=None):
    if os.path.exists(dest):
        logging.debug('Removing existing directory: %s' % dest)
        shutil.rmtree(dest)
    logging.info('Syncing directory: %s -> %s.' % (src, dest))

    shutil.copytree(src, dest, ignore=get_filter(opts))
    ensure_init(dest)


def sync(src, dest, module, opts=None):

    # Sync charmhelpers/__init__.py for bootstrap code.
    sync_pyfile(_src_path(src, '__init__'), dest)

    # Sync other __init__.py files in the path leading to module.
    m = []
    steps = module.split('.')[:-1]
    while steps:
        m.append(steps.pop(0))
        init = '.'.join(m + ['__init__'])
        sync_pyfile(_src_path(src, init),
                    os.path.dirname(_dest_path(dest, init)))

    # Sync the module, or maybe a .py file.
    if os.path.isdir(_src_path(src, module)):
        sync_directory(_src_path(src, module), _dest_path(dest, module), opts)
    elif _is_pyfile(_src_path(src, module)):
        sync_pyfile(_src_path(src, module),
                    os.path.dirname(_dest_path(dest, module)))
    else:
        logging.warn('Could not sync: %s. Neither a pyfile or directory, '
                     'does it even exist?' % module)


def parse_sync_options(options):
    if not options:
        return []
    return options.split(',')


def extract_options(inc, global_options=None):
    global_options = global_options or []
    if global_options and isinstance(global_options, str):
        global_options = [global_options]
    if '|' not in inc:
        return (inc, global_options)
    inc, opts = inc.split('|')
    return (inc, parse_sync_options(opts) + global_options)


def sync_helpers(include, src, dest, options=None):
    if os.path.exists(dest):
        logging.debug('Removing existing directory: %s' % dest)
        shutil.rmtree(dest)
    if not os.path.isdir(dest):
        os.makedirs(dest)

    global_options = parse_sync_options(options)

    for inc in include:
        if isinstance(inc, str):
            inc, opts = extract_options(inc, global_options)
            sync(src, dest, inc, opts)
        elif isinstance(inc, dict):
            # could also do nested dicts here.
            for k, v in inc.items():
                if isinstance(v, list):
                    for m in v:
                        inc, opts = extract_options(m, global_options)
                        sync(src, dest, '%s.%s' % (k, inc), opts)


if __name__ == '__main__':
    parser = optparse.OptionParser()
    parser.add_option('-c', '--config', action='store', dest='config',
                      default=None, help='helper config file')
    parser.add_option('-D', '--debug', action='store_true', dest='debug',
                      default=False, help='debug')
    parser.add_option('-r', '--repository', action='store', dest='repo',
                      help='charm-helpers git repository (overrides config)')
    parser.add_option('-d', '--destination', action='store', dest='dest_dir',
                      help='sync destination dir (overrides config)')
    (opts, args) = parser.parse_args()

    if opts.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    if opts.config:
        logging.info('Loading charm helper config from %s.' % opts.config)
        config = parse_config(opts.config)
        if not config:
            logging.error('Could not parse config from %s.' % opts.config)
            sys.exit(1)
    else:
        config = {}

    if 'repo' not in config:
        config['repo'] = CHARM_HELPERS_REPO
    if opts.repo:
        config['repo'] = opts.repo
    if opts.dest_dir:
        config['destination'] = opts.dest_dir

    if 'destination' not in config:
        logging.error('No destination dir. specified as option or config.')
        sys.exit(1)

    if 'include' not in config:
        if not args:
            logging.error('No modules to sync specified as option or config.')
            sys.exit(1)
        config['include'] = []
        [config['include'].append(a) for a in args]

    sync_options = None
    if 'options' in config:
        sync_options = config['options']
    tmpd = tempfile.mkdtemp()
    try:
        checkout = clone_helpers(tmpd, config['repo'])
        sync_helpers(config['include'], checkout, config['destination'],
                     options=sync_options)
    except Exception as e:
        logging.error("Could not sync: %s" % e)
        raise e
    finally:
        logging.debug('Cleaning up %s' % tmpd)
        shutil.rmtree(tmpd)



import subprocess
import os


def persistent_modprobe(module):
    """Load a kernel module and configure for auto-load on reboot."""
    if not os.path.exists('/etc/rc.modules'):
        open('/etc/rc.modules', 'a')
        os.chmod('/etc/rc.modules', 111)
    with open('/etc/rc.modules', 'r+') as modules:
        if module not in modules.read():
            modules.write('modprobe %s\n' % module)


def update_initramfs(version='all'):
    """Updates an initramfs image."""
    return subprocess.check_call(["dracut", "-f", version])



# Copyright 2016 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tests.contrib.hardware.test_utils import CharmTestCase
from tests.contrib.hardware.test_utils import patch_open

from tests.contrib.hardware.test_pci_helper import check_device
from tests.contrib.hardware.test_pci_helper import mocked_filehandle
from tests.contrib.hardware.test_pci_helper import mocked_globs
from tests.contrib.hardware.test_pci_helper import mocked_islink
from tests.contrib.hardware.test_pci_helper import mocked_listdir
from tests.contrib.hardware.test_pci_helper import mocked_realpath
from tests.contrib.hardware.test_pci_helper import mocked_subprocess

from mock import call
from mock import MagicMock
from mock import patch

from charmhelpers.contrib.hardware import pci

TO_PATCH = ["glob", "subprocess"]
NOT_JSON = "Im not json"


class PCITest(CharmTestCase):
    def setUp(self):
        super(PCITest, self).setUp(pci, TO_PATCH)

    def test_format_pci_addr(self):
        self.assertEqual(pci.format_pci_addr("0:0:1.1"), "0000:00:01.1")
        self.assertEqual(pci.format_pci_addr("0000:00:02.1"), "0000:00:02.1")


class PCINetDeviceTest(CharmTestCase):
    def setUp(self):
        super(PCINetDeviceTest, self).setUp(pci, TO_PATCH)

    @patch("os.path.islink")
    @patch("os.path.realpath")
    def eth_int(self, pci_address, _osrealpath, _osislink, subproc_map=None):
        self.glob.glob.side_effect = mocked_globs
        _osislink.side_effect = mocked_islink
        _osrealpath.side_effect = mocked_realpath
        self.subprocess.check_output.side_effect = mocked_subprocess(
            subproc_map=subproc_map
        )

        with patch_open() as (_open, _file):
            super_fh = mocked_filehandle()
            _file.readlines = MagicMock()
            _open.side_effect = super_fh._setfilename
            _file.read.side_effect = super_fh._getfilecontents_read
            _file.readlines.side_effect = super_fh._getfilecontents_readlines
            netint = pci.PCINetDevice(pci_address)
        return netint

    @patch("os.listdir")
    def test_base_eth_device(self, _oslistdir):
        _oslistdir.side_effect = mocked_listdir
        net = self.eth_int("0000:10:00.0")
        expect = {
            "interface_name": "eth2",
            "mac_address": "a8:9d:21:cf:93:fc",
            "pci_address": "0000:10:00.0",
            "state": "up",
            "sriov": False,
        }
        self.assertTrue(check_device(net, expect))

    @patch("charmhelpers.contrib.hardware.pci.get_sysnet_interfaces_and_macs")
    @patch("charmhelpers.contrib.hardware.pci.PCINetDevice.update_attributes")
    def test_update_interface_info(self, _update, _sysnet_ints):
        dev = pci.PCINetDevice("0000:10:00.0")
        _sysnet_ints.return_value = [
            {
                "interface": "eth2",
                "mac_address": "a8:9d:21:cf:93:fc",
                "pci_address": "0000:10:00.0",
                "state": "up",
                "sriov": False,
            },
            {
                "interface": "eth3",
                "mac_address": "a8:9d:21:cf:93:fd",
                "pci_address": "0000:10:00.1",
                "state": "down",
                "sriov": False,
            },
        ]
        dev.update_interface_info()
        self.assertEqual(dev.interface_name, "eth2")

    @patch("os.path.islink")
    @patch("os.path.realpath")
    @patch("charmhelpers.contrib.hardware.pci.is_sriov")
    @patch("charmhelpers.contrib.hardware.pci.get_sysnet_device_state")
    @patch("charmhelpers.contrib.hardware.pci.get_sysnet_mac")
    @patch("charmhelpers.contrib.hardware.pci.get_sysnet_interface")
    @patch("charmhelpers.contrib.hardware.pci.PCINetDevice.update_attributes")
    def test_get_sysnet_interfaces_and_macs(
        self, _update, _interface, _mac, _state, _sriov, _osrealpath, _osislink
    ):
        self.glob.glob.side_effect = (
            ["/sys/bus/pci/devices/0000:10:00.0"],
            [])
        _interface.return_value = "eth2"
        _mac.return_value = "a8:9d:21:cf:93:fc"
        _state.return_value = "up"
        _sriov.return_value = False
        _osrealpath.return_value = (
            "/sys/devices/pci0000:00/0000:00:02.0/"
            "0000:02:00.0/0000:03:00.0/0000:04:00.0/"
            "0000:05:01.0/0000:07:00.0"
        )
        expect = {
            "interface": "eth2",
            "mac_address": "a8:9d:21:cf:93:fc",
            "pci_address": "0000:07:00.0",
            "state": "up",
            "sriov": False,
        }
        self.assertEqual(pci.get_sysnet_interfaces_and_macs(), [expect])

    @patch("os.path.islink")
    @patch("os.path.realpath")
    @patch("charmhelpers.contrib.hardware.pci.is_sriov")
    @patch("charmhelpers.contrib.hardware.pci.get_sysnet_device_state")
    @patch("charmhelpers.contrib.hardware.pci.get_sysnet_mac")
    @patch("charmhelpers.contrib.hardware.pci.get_sysnet_interface")
    @patch("charmhelpers.contrib.hardware.pci.PCINetDevice.update_attributes")
    def test_get_sysnet_interfaces_and_macs_virtio(
        self, _update, _interface, _mac, _state, _sriov, _osrealpath, _osislink
    ):
        self.glob.glob.side_effect = (
            ["/sys/bus/pci/devices/0000:10:00.0"],
            [])
        _interface.return_value = "eth2"
        _mac.return_value = "a8:9d:21:cf:93:fc"
        _state.return_value = "up"
        _sriov.return_value = False
        _osrealpath.return_value = (
            "/sys/devices/pci0000:00/0000:00:07.0/" "virtio5"
        )
        expect = {
            "interface": "eth2",
            "mac_address": "a8:9d:21:cf:93:fc",
            "pci_address": "0000:00:07.0",
            "state": "up",
            "sriov": False,
        }
        self.assertEqual(pci.get_sysnet_interfaces_and_macs(), [expect])

    @patch("charmhelpers.contrib.hardware.pci.PCINetDevice.update_attributes")
    def test_get_sysnet_mac(self, _update):
        with patch_open() as (_open, _file):
            super_fh = mocked_filehandle()
            _file.readlines = MagicMock()
            _open.side_effect = super_fh._setfilename
            _file.read.side_effect = super_fh._getfilecontents_read
            macaddr = pci.get_sysnet_mac(
                "/sys/bus/pci/devices/0000:10:00.1", "eth3")
        self.assertEqual(macaddr, "a8:9d:21:cf:93:fd")

    @patch("charmhelpers.contrib.hardware.pci.PCINetDevice.update_attributes")
    def test_get_sysnet_device_state(self, _update):
        with patch_open() as (_open, _file):
            super_fh = mocked_filehandle()
            _file.readlines = MagicMock()
            _open.side_effect = super_fh._setfilename
            _file.read.side_effect = super_fh._getfilecontents_read
            state = pci.get_sysnet_device_state(
                "/sys/bus/pci/devices/0000:10:00.1", "eth3")
        self.assertEqual(state, "down")

    @patch("os.listdir")
    @patch("charmhelpers.contrib.hardware.pci.PCINetDevice.update_attributes")
    def test_get_sysnet_interface(self, _update, _oslistdir):
        _oslistdir.side_effect = mocked_listdir
        with patch_open() as (_open, _file):
            super_fh = mocked_filehandle()
            _file.readlines = MagicMock()
            _open.side_effect = super_fh._setfilename
            _file.read.side_effect = super_fh._getfilecontents_read
            self.assertEqual(
                pci.get_sysnet_interface("/sys/bus/pci/devices/0000:10:00.1"),
                "eth3"
            )

    @patch("charmhelpers.contrib.hardware.pci.get_sysnet_interfaces_and_macs")
    def test__set_sriov_numvfs(self, mock_sysnet_ints):
        mock_sysnet_ints.side_effect = (
            [
                {
                    "interface": "eth2",
                    "mac_address": "a8:9d:21:cf:93:fc",
                    "pci_address": "0000:10:00.0",
                    "state": "up",
                    "sriov": True,
                    "sriov_totalvfs": 7,
                    "sriov_numvfs": 0,
                }
            ],
            [
                {
                    "interface": "eth2",
                    "mac_address": "a8:9d:21:cf:93:fc",
                    "pci_address": "0000:10:00.0",
                    "state": "up",
                    "sriov": True,
                    "sriov_totalvfs": 7,
                    "sriov_numvfs": 4,
                }
            ],
        )
        dev = pci.PCINetDevice("0000:10:00.0")
        self.assertEqual("eth2", dev.interface_name)
        self.assertTrue(dev.sriov)
        self.assertEqual(7, dev.sriov_totalvfs)
        self.assertEqual(0, dev.sriov_numvfs)

        with patch_open() as (mock_open, mock_file):
            dev._set_sriov_numvfs(4)
            mock_open.assert_called_with(
                "/sys/bus/pci/devices/0000:10:00.0/sriov_numvfs", "w"
            )
            mock_file.write.assert_called_with("4")
            self.assertTrue(dev.sriov)
            self.assertEqual(7, dev.sriov_totalvfs)
            self.assertEqual(4, dev.sriov_numvfs)

    @patch("charmhelpers.contrib.hardware.pci.PCINetDevice._set_sriov_numvfs")
    def test_set_sriov_numvfs(self, mock__set_sriov_numvfs):
        dev = pci.PCINetDevice("0000:10:00.0")
        dev.sriov = True
        dev.set_sriov_numvfs(4)
        mock__set_sriov_numvfs.assert_has_calls([call(0), call(4)])

    @patch("charmhelpers.contrib.hardware.pci.PCINetDevice._set_sriov_numvfs")
    def test_set_sriov_numvfs_avoid_call(self, mock__set_sriov_numvfs):
        dev = pci.PCINetDevice("0000:10:00.0")
        dev.sriov = True
        dev.sriov_numvfs = 4
        dev.set_sriov_numvfs(4)
        self.assertFalse(mock__set_sriov_numvfs.called)


class PCINetDevicesTest(CharmTestCase):
    def setUp(self):
        super(PCINetDevicesTest, self).setUp(pci, TO_PATCH)

    @patch("os.path.islink")
    def pci_devs(self, _osislink, subproc_map=None):
        self.glob.glob.side_effect = mocked_globs
        rp_patcher = patch("os.path.realpath")
        rp_mock = rp_patcher.start()
        rp_mock.side_effect = mocked_realpath
        _osislink.side_effect = mocked_islink
        self.subprocess.check_output.side_effect = mocked_subprocess(
            subproc_map=subproc_map
        )
        listdir_patch = patch("os.listdir")
        listdir_mock = listdir_patch.start()
        listdir_mock.side_effect = mocked_listdir

        with patch_open() as (_open, _file):
            super_fh = mocked_filehandle()
            _file.readlines = MagicMock()
            _open.side_effect = super_fh._setfilename
            _file.read.side_effect = super_fh._getfilecontents_read
            _file.readlines.side_effect = super_fh._getfilecontents_readlines
            devices = pci.PCINetDevices()
        rp_patcher.stop()
        listdir_patch.stop()
        return devices

    def test_base(self):
        devices = self.pci_devs()
        self.assertEqual(len(devices.pci_devices), 2)
        expect = {
            "0000:10:00.0": {
                "interface_name": "eth2",
                "mac_address": "a8:9d:21:cf:93:fc",
                "pci_address": "0000:10:00.0",
                "state": "up",
                "sriov": False,
            },
            "0000:10:00.1": {
                "interface_name": "eth3",
                "mac_address": "a8:9d:21:cf:93:fd",
                "pci_address": "0000:10:00.1",
                "state": "down",
                "sriov": False,
            },
        }
        for device in devices.pci_devices:
            self.assertTrue(check_device(device, expect[device.pci_address]))

    def test_get_pci_ethernet_addresses(self):
        self.pci_devs()
        expect = ["0000:10:00.0", "0000:10:00.1"]
        self.assertEqual(pci.get_pci_ethernet_addresses(), expect)

    @patch("charmhelpers.contrib.hardware.pci.PCINetDevice.update_attributes")
    def test_update_devices(self, _update):
        devices = self.pci_devs()
        call_count = _update.call_count
        devices.update_devices()
        self.assertEqual(_update.call_count, call_count + 2)

    def test_get_macs(self):
        devices = self.pci_devs()
        expect = ["a8:9d:21:cf:93:fc", "a8:9d:21:cf:93:fd"]
        self.assertEqual(devices.get_macs(), expect)

    def test_get_device_from_mac(self):
        devices = self.pci_devs()
        expect = {
            "0000:10:00.1": {
                "interface_name": "eth3",
                "mac_address": "a8:9d:21:cf:93:fd",
                "pci_address": "0000:10:00.1",
                "state": "down",
                "sriov": False,
            }
        }
        self.assertTrue(
            check_device(
                devices.get_device_from_mac("a8:9d:21:cf:93:fd"),
                expect["0000:10:00.1"],
            )
        )

    def test_get_device_from_pci_address(self):
        devices = self.pci_devs()
        expect = {
            "0000:10:00.1": {
                "interface_name": "eth3",
                "mac_address": "a8:9d:21:cf:93:fd",
                "pci_address": "0000:10:00.1",
                "state": "down",
            }
        }
        self.assertTrue(
            check_device(
                devices.get_device_from_pci_address("0000:10:00.1"),
                expect["0000:10:00.1"],
            )
        )



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from mock import patch

from charmhelpers.contrib.hardening.host.checks import minimize_access


class MinimizeAccessTestCase(TestCase):

    @patch.object(minimize_access.utils, 'get_settings', lambda x:
                  {'environment': {'extra_user_paths': []},
                   'security': {'users_allow': []}})
    def test_default_options(self):
        audits = minimize_access.get_audits()
        self.assertEqual(3, len(audits))

        # First audit is to ensure that all folders in the $PATH variable
        # are read-only.
        self.assertTrue(isinstance(audits[0], minimize_access.ReadOnly))
        self.assertEqual({'/usr/local/sbin', '/usr/local/bin',
                          '/usr/sbin', '/usr/bin', '/bin'}, audits[0].paths)

        # Second audit is to ensure that the /etc/shadow is only readable
        # by the root user.
        self.assertTrue(isinstance(audits[1],
                                   minimize_access.FilePermissionAudit))
        self.assertEqual(audits[1].paths[0], '/etc/shadow')
        self.assertEqual(audits[1].mode, 0o0600)

        # Last audit is to ensure that only root has access to the su
        self.assertTrue(isinstance(audits[2],
                                   minimize_access.FilePermissionAudit))
        self.assertEqual(audits[2].paths[0], '/bin/su')
        self.assertEqual(audits[2].mode, 0o0750)

    @patch.object(minimize_access.utils, 'get_settings', lambda x:
                  {'environment': {'extra_user_paths': []},
                   'security': {'users_allow': ['change_user']}})
    def test_allow_change_user(self):
        audits = minimize_access.get_audits()
        self.assertEqual(2, len(audits))

        # First audit is to ensure that all folders in the $PATH variable
        # are read-only.
        self.assertTrue(isinstance(audits[0], minimize_access.ReadOnly))
        self.assertEqual({'/usr/local/sbin', '/usr/local/bin',
                          '/usr/sbin', '/usr/bin', '/bin'}, audits[0].paths)

        # Second audit is to ensure that the /etc/shadow is only readable
        # by the root user.
        self.assertTrue(isinstance(audits[1],
                                   minimize_access.FilePermissionAudit))
        self.assertEqual(audits[1].paths[0], '/etc/shadow')
        self.assertEqual(audits[1].mode, 0o0600)



# Copyright 2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
A Pythonic API to interact with the charm hook environment.

:author: Stuart Bishop <stuart.bishop@canonical.com>
'''

from charmhelpers.core import hookenv

from collections import OrderedDict, UserDict


class Relations(OrderedDict):
    '''Mapping relation name -> relation id -> Relation.

    >>> rels = Relations()
    >>> rels['sprog']['sprog:12']['client/6']['widget']
    'remote widget'
    >>> rels['sprog']['sprog:12'].local['widget'] = 'local widget'
    >>> rels['sprog']['sprog:12'].local['widget']
    'local widget'
    >>> rels.peer.local['widget']
    'local widget on the peer relation'
    '''
    def __init__(self):
        super(Relations, self).__init__()
        for relname in sorted(hookenv.relation_types()):
            self[relname] = OrderedDict()
            relids = hookenv.relation_ids(relname)
            relids.sort(key=lambda x: int(x.split(':', 1)[-1]))
            for relid in relids:
                self[relname][relid] = Relation(relid)

    @property
    def peer(self):
        peer_relid = hookenv.peer_relation_id()
        for rels in self.values():
            if peer_relid in rels:
                return rels[peer_relid]


class Relation(OrderedDict):
    '''Mapping of unit -> remote RelationInfo for a relation.

    This is an OrderedDict mapping, ordered numerically by
    by unit number.

    Also provides access to the local RelationInfo, and peer RelationInfo
    instances by the 'local' and 'peers' attributes.

    >>> r = Relation('sprog:12')
    >>> r.keys()
    ['client/9', 'client/10']     # Ordered numerically
    >>> r['client/10']['widget']  # A remote RelationInfo setting
    'remote widget'
    >>> r.local['widget']         # The local RelationInfo setting
    'local widget'
    '''
    relid = None    # The relation id.
    relname = None  # The relation name (also known as relation type).
    service = None  # The remote service name, if known.
    local = None    # The local end's RelationInfo.
    peers = None    # Map of peer -> RelationInfo. None if no peer relation.

    def __init__(self, relid):
        remote_units = hookenv.related_units(relid)
        remote_units.sort(key=lambda u: int(u.split('/', 1)[-1]))
        super(Relation, self).__init__((unit, RelationInfo(relid, unit))
                                       for unit in remote_units)

        self.relname = relid.split(':', 1)[0]
        self.relid = relid
        self.local = RelationInfo(relid, hookenv.local_unit())

        for relinfo in self.values():
            self.service = relinfo.service
            break

        # If we have peers, and they have joined both the provided peer
        # relation and this relation, we can peek at their data too.
        # This is useful for creating consensus without leadership.
        peer_relid = hookenv.peer_relation_id()
        if peer_relid and peer_relid != relid:
            peers = hookenv.related_units(peer_relid)
            if peers:
                peers.sort(key=lambda u: int(u.split('/', 1)[-1]))
                self.peers = OrderedDict((peer, RelationInfo(relid, peer))
                                         for peer in peers)
            else:
                self.peers = OrderedDict()
        else:
            self.peers = None

    def __str__(self):
        return '{} ({})'.format(self.relid, self.service)


class RelationInfo(UserDict):
    '''The bag of data at an end of a relation.

    Every unit participating in a relation has a single bag of
    data associated with that relation. This is that bag.

    The bag of data for the local unit may be updated. Remote data
    is immutable and will remain static for the duration of the hook.

    Changes made to the local units relation data only become visible
    to other units after the hook completes successfully. If the hook
    does not complete successfully, the changes are rolled back.

    Unlike standard Python mappings, setting an item to None is the
    same as deleting it.

    >>> relinfo = RelationInfo('db:12')  # Default is the local unit.
    >>> relinfo['user'] = 'fred'
    >>> relinfo['user']
    'fred'
    >>> relinfo['user'] = None
    >>> 'fred' in relinfo
    False

    This class wraps hookenv.relation_get and hookenv.relation_set.
    All caching is left up to these two methods to avoid synchronization
    issues. Data is only loaded on demand.
    '''
    relid = None    # The relation id.
    relname = None  # The relation name (also know as the relation type).
    unit = None     # The unit id.
    number = None   # The unit number (integer).
    service = None  # The service name.

    def __init__(self, relid, unit):
        self.relname = relid.split(':', 1)[0]
        self.relid = relid
        self.unit = unit
        self.service, num = self.unit.split('/', 1)
        self.number = int(num)

    def __str__(self):
        return '{} ({})'.format(self.relid, self.unit)

    @property
    def data(self):
        return hookenv.relation_get(rid=self.relid, unit=self.unit)

    def __setitem__(self, key, value):
        if self.unit != hookenv.local_unit():
            raise TypeError('Attempting to set {} on remote unit {}'
                            ''.format(key, self.unit))
        if value is not None and not isinstance(value, str):
            # We don't do implicit casting. This would cause simple
            # types like integers to be read back as strings in subsequent
            # hooks, and mutable types would require a lot of wrapping
            # to ensure relation-set gets called when they are mutated.
            raise ValueError('Only string values allowed')
        hookenv.relation_set(self.relid, {key: value})

    def __delitem__(self, key):
        # Deleting a key and setting it to null is the same thing in
        # Juju relations.
        self[key] = None


class Leader(UserDict):
    def __init__(self):
        pass  # Don't call superclass initializer, as it will nuke self.data

    @property
    def data(self):
        return hookenv.leader_get()

    def __setitem__(self, key, value):
        if not hookenv.is_leader():
            raise TypeError('Not the leader. Cannot change leader settings.')
        if value is not None and not isinstance(value, str):
            # We don't do implicit casting. This would cause simple
            # types like integers to be read back as strings in subsequent
            # hooks, and mutable types would require a lot of wrapping
            # to ensure leader-set gets called when they are mutated.
            raise ValueError('Only string values allowed')
        hookenv.leader_set({key: value})

    def __delitem__(self, key):
        # Deleting a key and setting it to null is the same thing in
        # Juju leadership settings.
        self[key] = None



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from charmhelpers.contrib.hardening.audits import BaseAudit


class BaseAuditTestCase(TestCase):

    def setUp(self):
        super(BaseAuditTestCase, self).setUp()

    def test_take_action_default(self):
        check = BaseAudit()
        take_action = check._take_action()
        self.assertTrue(take_action)

    def test_take_action_unless_true(self):
        check = BaseAudit(unless=True)
        take_action = check._take_action()
        self.assertFalse(take_action)

    def test_take_action_unless_false(self):
        check = BaseAudit(unless=False)
        take_action = check._take_action()
        self.assertTrue(take_action)

    def test_take_action_unless_callback_false(self):
        def callback():
            return False
        check = BaseAudit(unless=callback)
        take_action = check._take_action()
        self.assertTrue(take_action)

    def test_take_action_unless_callback_true(self):
        def callback():
            return True
        check = BaseAudit(unless=callback)
        take_action = check._take_action()
        self.assertFalse(take_action)



import unittest
import subprocess

from mock import patch

import charmhelpers.contrib.storage.linux.lvm as lvm

PVDISPLAY = b"""
  --- Physical volume ---
  PV Name               /dev/loop0
  VG Name               foo
  PV Size               10.00 MiB / not usable 2.00 MiB
  Allocatable           yes
  PE Size               4.00 MiB
  Total PE              2
  Free PE               2
  Allocated PE          0
  PV UUID               fyVqlr-pyrL-89On-f6MD-U91T-dEfc-SL0V2V

"""

EMPTY_VG_IN_PVDISPLAY = b"""
  --- Physical volume ---
  PV Name               /dev/loop0
  VG Name
  PV Size               10.00 MiB / not usable 2.00 MiB
  Allocatable           yes
  PE Size               4.00 MiB
  Total PE              2
  Free PE               2
  Allocated PE          0
  PV UUID               fyVqlr-pyrL-89On-f6MD-U91T-dEfc-SL0V2V

"""
LVS_DEFAULT = b"""
  cinder-volumes-pool
  testvol
  volume-48be6ba0-84c3-4b8d-9be5-e68e47fc7682
  volume-f8c1d2fd-1fa1-4d84-b4e0-431dba7d582e
"""
LVS_WITH_VG = b"""
  cinder-volumes cinder-volumes-pool
  cinder-volumes testvol
  cinder-volumes volume-48be6ba0-84c3-4b8d-9be5-e68e47fc7682
  cinder-volumes volume-f8c1d2fd-1fa1-4d84-b4e0-431dba7d582e
"""
LVS_THIN_POOLS = b"""
  cinder-volumes-pool
"""
LVS_THIN_POOLS_WITH_VG = b"""
  cinder-volumes cinder-volumes-pool
"""

# It's a mouthful.
STORAGE_LINUX_LVM = 'charmhelpers.contrib.storage.linux.lvm'


class LVMStorageUtilsTests(unittest.TestCase):
    def test_find_volume_group_on_pv(self):
        """It determines any volume group assigned to a LVM PV"""
        with patch(STORAGE_LINUX_LVM + '.check_output') as check_output:
            check_output.return_value = PVDISPLAY
            vg = lvm.list_lvm_volume_group('/dev/loop0')
            self.assertEquals(vg, 'foo')

    def test_find_empty_volume_group_on_pv(self):
        """Return empty string when no volume group is assigned to the PV"""
        with patch(STORAGE_LINUX_LVM + '.check_output') as check_output:
            check_output.return_value = EMPTY_VG_IN_PVDISPLAY
            vg = lvm.list_lvm_volume_group('/dev/loop0')
            self.assertEquals(vg, '')

    @patch(STORAGE_LINUX_LVM + '.list_lvm_volume_group')
    def test_deactivate_lvm_volume_groups(self, ls_vg):
        """It deactivates active volume groups on LVM PV"""
        ls_vg.return_value = 'foo-vg'
        with patch(STORAGE_LINUX_LVM + '.check_call') as check_call:
            lvm.deactivate_lvm_volume_group('/dev/loop0')
            check_call.assert_called_with(['vgchange', '-an', 'foo-vg'])

    def test_remove_lvm_physical_volume(self):
        """It removes LVM physical volume signatures from block device"""
        with patch(STORAGE_LINUX_LVM + '.Popen') as popen:
            lvm.remove_lvm_physical_volume('/dev/foo')
            popen.assert_called_with(['pvremove', '-ff', '/dev/foo'], stdin=-1)

    def test_is_physical_volume(self):
        """It properly reports block dev is an LVM PV"""
        with patch(STORAGE_LINUX_LVM + '.check_output') as check_output:
            check_output.return_value = PVDISPLAY
            self.assertTrue(lvm.is_lvm_physical_volume('/dev/loop0'))

    def test_is_not_physical_volume(self):
        """It properly reports block dev is an LVM PV"""
        with patch(STORAGE_LINUX_LVM + '.check_output') as check_output:
            check_output.side_effect = subprocess.CalledProcessError('cmd', 2)
            self.assertFalse(lvm.is_lvm_physical_volume('/dev/loop0'))

    def test_pvcreate(self):
        """It correctly calls pvcreate for a given block dev"""
        with patch(STORAGE_LINUX_LVM + '.check_call') as check_call:
            lvm.create_lvm_physical_volume('/dev/foo')
            check_call.assert_called_with(['pvcreate', '/dev/foo'])

    def test_vgcreate(self):
        """It correctly calls vgcreate for given block dev and vol group"""
        with patch(STORAGE_LINUX_LVM + '.check_call') as check_call:
            lvm.create_lvm_volume_group('foo-vg', '/dev/foo')
            check_call.assert_called_with(['vgcreate', 'foo-vg', '/dev/foo'])

    def test_list_logical_volumes(self):
        with patch(STORAGE_LINUX_LVM + '.check_output') as check_output:
            check_output.return_value = LVS_DEFAULT
            self.assertEqual(lvm.list_logical_volumes(), [
                'cinder-volumes-pool',
                'testvol',
                'volume-48be6ba0-84c3-4b8d-9be5-e68e47fc7682',
                'volume-f8c1d2fd-1fa1-4d84-b4e0-431dba7d582e'])
            check_output.assert_called_with([
                'lvs',
                '--options',
                'lv_name',
                '--noheadings'])

    def test_list_logical_volumes_empty(self):
        with patch(STORAGE_LINUX_LVM + '.check_output') as check_output:
            check_output.return_value = b''
            self.assertEqual(lvm.list_logical_volumes(), [])

    def test_list_logical_volumes_path_mode(self):
        with patch(STORAGE_LINUX_LVM + '.check_output') as check_output:
            check_output.return_value = LVS_WITH_VG
            self.assertEqual(lvm.list_logical_volumes(path_mode=True), [
                'cinder-volumes/cinder-volumes-pool',
                'cinder-volumes/testvol',
                'cinder-volumes/volume-48be6ba0-84c3-4b8d-9be5-e68e47fc7682',
                'cinder-volumes/volume-f8c1d2fd-1fa1-4d84-b4e0-431dba7d582e'])
            check_output.assert_called_with([
                'lvs',
                '--options',
                'vg_name,lv_name',
                '--noheadings'])

    def test_list_logical_volumes_select_criteria(self):
        with patch(STORAGE_LINUX_LVM + '.check_output') as check_output:
            check_output.return_value = LVS_THIN_POOLS
            self.assertEqual(
                lvm.list_logical_volumes(select_criteria='lv_attr =~ ^t'),
                ['cinder-volumes-pool'])
            check_output.assert_called_with([
                'lvs',
                '--options',
                'lv_name',
                '--noheadings',
                '--select',
                'lv_attr =~ ^t'])

    def test_list_thin_logical_volume_pools(self):
        with patch(STORAGE_LINUX_LVM + '.check_output') as check_output:
            check_output.return_value = LVS_THIN_POOLS
            self.assertEqual(
                lvm.list_thin_logical_volume_pools(),
                ['cinder-volumes-pool'])
            check_output.assert_called_with([
                'lvs',
                '--options',
                'lv_name',
                '--noheadings',
                '--select',
                'lv_attr =~ ^t'])

    def test_list_thin_logical_volume_pools_path_mode(self):
        with patch(STORAGE_LINUX_LVM + '.check_output') as check_output:
            check_output.return_value = LVS_THIN_POOLS_WITH_VG
            self.assertEqual(
                lvm.list_thin_logical_volume_pools(path_mode=True),
                ['cinder-volumes/cinder-volumes-pool'])
            check_output.assert_called_with([
                'lvs',
                '--options',
                'vg_name,lv_name',
                '--noheadings',
                '--select',
                'lv_attr =~ ^t'])

    def test_extend_logical_volume_by_device(self):
        """It correctly calls pvcreate for a given block dev"""
        with patch(STORAGE_LINUX_LVM + '.check_call') as check_call:
            lvm.extend_logical_volume_by_device('mylv', '/dev/foo')
            check_call.assert_called_with(['lvextend', 'mylv', '/dev/foo'])

    def test_create_logical_volume_nosize(self):
        with patch(STORAGE_LINUX_LVM + '.check_call') as check_call:
            lvm.create_logical_volume('testlv', 'testvg')
            check_call.assert_called_with([
                'lvcreate',
                '--yes',
                '-l',
                '100%FREE',
                '-n', 'testlv', 'testvg'
            ])

    def test_create_logical_volume_size(self):
        with patch(STORAGE_LINUX_LVM + '.check_call') as check_call:
            lvm.create_logical_volume('testlv', 'testvg', '10G')
            check_call.assert_called_with([
                'lvcreate',
                '--yes',
                '-L',
                '10G',
                '-n', 'testlv', 'testvg'
            ])



import subprocess
import os

from tests.helpers import patch_open
from testtools import TestCase
from mock import (
    patch,
    MagicMock,
    call,
)
from charmhelpers.fetch import centos as fetch


def getenv(update=None):
    # return a copy of os.environ with update applied.
    # this was necessary because some modules modify os.environment directly
    copy = os.environ.copy()
    if update is not None:
        copy.update(update)
    return copy


class FetchTest(TestCase):

    @patch("charmhelpers.fetch.log")
    @patch('yum.YumBase.doPackageLists')
    def test_filter_packages_missing_centos(self, yumBase, log):

        class MockPackage:
            def __init__(self, name):
                self.base_package_name = name

        yum_dict = {
            'installed': {
                MockPackage('vim')
            },
            'available': {
                MockPackage('vim')
            }
        }
        import yum
        yum.YumBase.return_value.doPackageLists.return_value = yum_dict
        result = fetch.filter_installed_packages(['vim', 'emacs'])
        self.assertEquals(result, ['emacs'])

    @patch("charmhelpers.fetch.log")
    def test_filter_packages_none_missing_centos(self, log):

        class MockPackage:
            def __init__(self, name):
                self.base_package_name = name

        yum_dict = {
            'installed': {
                MockPackage('vim')
            },
            'available': {
                MockPackage('vim')
            }
        }
        import yum
        yum.yumBase.return_value.doPackageLists.return_value = yum_dict
        result = fetch.filter_installed_packages(['vim'])
        self.assertEquals(result, [])

    @patch('charmhelpers.fetch.centos.log')
    @patch('yum.YumBase.doPackageLists')
    def test_filter_packages_not_available_centos(self, yumBase, log):

        class MockPackage:
            def __init__(self, name):
                self.base_package_name = name

        yum_dict = {
            'installed': {
                MockPackage('vim')
            }
        }
        import yum
        yum.YumBase.return_value.doPackageLists.return_value = yum_dict

        result = fetch.filter_installed_packages(['vim', 'joe'])
        self.assertEquals(result, ['joe'])

    @patch('charmhelpers.fetch.centos.log')
    def test_add_source_none_centos(self, log):
        fetch.add_source(source=None)
        self.assertTrue(log.called)

    @patch('charmhelpers.fetch.centos.log')
    @patch('os.listdir')
    def test_add_source_http_centos(self, listdir, log):
        source = "http://archive.ubuntu.com/ubuntu raring-backports main"
        with patch_open() as (mock_open, mock_file):
            fetch.add_source(source=source)
            listdir.assert_called_with('/etc/yum.repos.d/')
            mock_file.write.assert_has_calls([
                call("[archive.ubuntu.com_ubuntu raring-backports main]\n"),
                call("name=archive.ubuntu.com/ubuntu raring-backports main\n"),
                call("baseurl=http://archive.ubuntu.com/ubuntu raring"
                     "-backports main\n\n")])

    @patch('charmhelpers.fetch.centos.log')
    @patch('os.listdir')
    @patch('subprocess.check_call')
    def test_add_source_http_and_key_id_centos(self, check_call,
                                               listdir, log):
        source = "http://archive.ubuntu.com/ubuntu raring-backports main"
        key_id = "akey"
        with patch_open() as (mock_open, mock_file):
            fetch.add_source(source=source, key=key_id)
            listdir.assert_called_with('/etc/yum.repos.d/')
            mock_file.write.assert_has_calls([
                call("[archive.ubuntu.com_ubuntu raring-backports main]\n"),
                call("name=archive.ubuntu.com/ubuntu raring-backports main\n"),
                call("baseurl=http://archive.ubuntu.com/ubuntu raring"
                     "-backports main\n\n")])
        check_call.assert_called_with(['rpm', '--import', key_id])

    @patch('charmhelpers.fetch.centos.log')
    @patch('os.listdir')
    @patch('subprocess.check_call')
    def test_add_source_https_and_key_id_centos(self, check_call,
                                                listdir, log):
        source = "https://USER:PASS@private-ppa.launchpad.net/project/awesome"
        key_id = "GPGPGP"
        with patch_open() as (mock_open, mock_file):
            fetch.add_source(source=source, key=key_id)
            listdir.assert_called_with('/etc/yum.repos.d/')
            mock_file.write.assert_has_calls([
                call("[_USER:PASS@private-ppa.launchpad"
                     ".net_project_awesome]\n"),
                call("name=/USER:PASS@private-ppa.launchpad.net"
                     "/project/awesome\n"),
                call("baseurl=https://USER:PASS@private-ppa.launchpad.net"
                     "/project/awesome\n\n")])
        check_call.assert_called_with(['rpm', '--import', key_id])

    @patch('charmhelpers.fetch.centos.log')
    @patch.object(fetch, 'NamedTemporaryFile')
    @patch('os.listdir')
    @patch('subprocess.check_call')
    def test_add_source_http_and_key_centos(self, check_call,
                                            listdir, temp_file, log):
        source = "http://archive.ubuntu.com/ubuntu raring-backports main"
        key = '''
            -----BEGIN PGP PUBLIC KEY BLOCK-----
            [...]
            -----END PGP PUBLIC KEY BLOCK-----
            '''
        file_mock = MagicMock()
        file_mock.name = 'temporary_file'
        temp_file.return_value.__enter__.return_value = file_mock
        listdir.return_value = []

        with patch_open() as (mock_open, mock_file):
            fetch.add_source(source=source, key=key)
            listdir.assert_called_with('/etc/yum.repos.d/')
            self.assertTrue(log.called)
        check_call.assert_called_with(['rpm', '--import', file_mock.name])
        file_mock.write.assert_called_once_with(key)
        file_mock.flush.assert_called_once_with()
        file_mock.seek.assert_called_once_with(0)


class YumTests(TestCase):

    @patch('subprocess.call')
    @patch('charmhelpers.fetch.centos.log')
    def test_yum_upgrade_non_fatal(self, log, mock_call):
        options = ['--foo', '--bar']
        fetch.upgrade(options)

        mock_call.assert_called_with(['yum', '--assumeyes',
                                      '--foo', '--bar', 'upgrade'],
                                     env=getenv())

    @patch('subprocess.check_call')
    @patch('charmhelpers.fetch.centos.log')
    def test_yum_upgrade_fatal(self, log, mock_call):
        options = ['--foo', '--bar']
        fetch.upgrade(options, fatal=True)

        mock_call.assert_called_with(['yum', '--assumeyes',
                                      '--foo', '--bar', 'upgrade'],
                                     env=getenv())

    @patch('subprocess.call')
    @patch('charmhelpers.fetch.centos.log')
    def test_installs_yum_packages(self, log, mock_call):
        packages = ['foo', 'bar']
        options = ['--foo', '--bar']

        fetch.install(packages, options)

        mock_call.assert_called_with(['yum', '--assumeyes',
                                      '--foo', '--bar', 'install',
                                      'foo', 'bar'],
                                     env=getenv())

    @patch('subprocess.call')
    @patch('charmhelpers.fetch.centos.log')
    def test_installs_yum_packages_without_options(self, log, mock_call):
        packages = ['foo', 'bar']
        fetch.install(packages)

        mock_call.assert_called_with(['yum', '--assumeyes',
                                      'install', 'foo', 'bar'],
                                     env=getenv())

    @patch('subprocess.call')
    @patch('charmhelpers.fetch.centos.log')
    def test_installs_yum_packages_as_string(self, log, mock_call):
        packages = 'foo bar'
        fetch.install(packages)

        mock_call.assert_called_with(['yum', '--assumeyes',
                                      'install', 'foo bar'],
                                     env=getenv())

    @patch('subprocess.check_call')
    @patch('charmhelpers.fetch.centos.log')
    def test_installs_yum_packages_with_possible_errors(self, log, mock_call):
        packages = ['foo', 'bar']
        options = ['--foo', '--bar']

        fetch.install(packages, options, fatal=True)

        mock_call.assert_called_with(['yum', '--assumeyes',
                                      '--foo', '--bar',
                                      'install', 'foo', 'bar'],
                                     env=getenv())

    @patch('subprocess.check_call')
    @patch('charmhelpers.fetch.centos.log')
    def test_purges_yum_packages_as_string_fatal(self, log, mock_call):
        packages = 'irrelevant names'
        mock_call.side_effect = OSError('fail')

        self.assertRaises(OSError, fetch.purge, packages, fatal=True)
        self.assertTrue(log.called)

    @patch('subprocess.check_call')
    @patch('charmhelpers.fetch.centos.log')
    def test_purges_yum_packages_fatal(self, log, mock_call):
        packages = ['irrelevant', 'names']
        mock_call.side_effect = OSError('fail')

        self.assertRaises(OSError, fetch.purge, packages, fatal=True)
        self.assertTrue(log.called)

    @patch('subprocess.call')
    @patch('charmhelpers.fetch.centos.log')
    def test_purges_yum_packages_as_string_nofatal(self, log, mock_call):
        packages = 'foo bar'
        fetch.purge(packages)

        self.assertTrue(log.called)
        mock_call.assert_called_with(['yum', '--assumeyes',
                                      'remove', 'foo bar'],
                                     env=getenv())

    @patch('subprocess.call')
    @patch('charmhelpers.fetch.centos.log')
    def test_purges_yum_packages_nofatal(self, log, mock_call):
        packages = ['foo', 'bar']
        fetch.purge(packages)

        self.assertTrue(log.called)
        mock_call.assert_called_with(['yum', '--assumeyes',
                                      'remove', 'foo', 'bar'],
                                     env=getenv())

    @patch('subprocess.check_call')
    @patch('charmhelpers.fetch.centos.log')
    def test_yum_update_fatal(self, log, check_call):
        fetch.update(fatal=True)
        check_call.assert_called_with(['yum', '--assumeyes', 'update'],
                                      env=getenv())
        self.assertTrue(log.called)

    @patch('subprocess.check_output')
    @patch('charmhelpers.fetch.centos.log')
    def test_yum_search(self, log, check_output):
        package = ['irrelevant']

        from charmhelpers.fetch.centos import yum_search
        yum_search(package)
        check_output.assert_called_with(['yum', 'search', 'irrelevant'])
        self.assertTrue(log.called)

    @patch('subprocess.check_call')
    @patch('time.sleep')
    def test_run_yum_command_retries_if_fatal(self, check_call, sleep):
        """The _run_yum_command function retries the command if it can't get
        the YUM lock."""
        self.called = False

        def side_effect(*args, **kwargs):
            """
            First, raise an exception (can't acquire lock), then return 0
            (the lock is grabbed).
            """
            if not self.called:
                self.called = True
                raise subprocess.CalledProcessError(
                    returncode=1, cmd="some command")
            else:
                return 0

        check_call.side_effect = side_effect
        check_call.return_value = 0
        from charmhelpers.fetch.centos import _run_yum_command
        _run_yum_command(["some", "command"], fatal=True)
        self.assertTrue(sleep.called)



# Copyright 2018-2021 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

import charmhelpers.contrib.openstack.alternatives as alternatives
import charmhelpers.contrib.openstack.context as context

import charmhelpers.core.hookenv as hookenv
import charmhelpers.core.host as host
import charmhelpers.core.templating as templating
import charmhelpers.core.unitdata as unitdata

VAULTLOCKER_BACKEND = 'charm-vaultlocker'


class VaultKVContext(context.OSContextGenerator):
    """Vault KV context for interaction with vault-kv interfaces"""
    interfaces = ['secrets-storage']

    def __init__(self, secret_backend=None):
        super(context.OSContextGenerator, self).__init__()
        self.secret_backend = (
            secret_backend or 'charm-{}'.format(hookenv.service_name())
        )

    def __call__(self):
        try:
            import hvac
        except ImportError:
            # BUG: #1862085 - if the relation is made to vault, but the
            # 'encrypt' option is not made, then the charm errors with an
            # import warning.  This catches that, logs a warning, and returns
            # with an empty context.
            hookenv.log("VaultKVContext: trying to use hvac pythong module "
                        "but it's not available.  Is secrets-stroage relation "
                        "made, but encrypt option not set?",
                        level=hookenv.WARNING)
            # return an empty context on hvac import error
            return {}
        ctxt = {}
        # NOTE(hopem): see https://bugs.launchpad.net/charm-helpers/+bug/1849323
        db = unitdata.kv()
        # currently known-good secret-id
        secret_id = db.get('secret-id')

        for relation_id in hookenv.relation_ids(self.interfaces[0]):
            for unit in hookenv.related_units(relation_id):
                data = hookenv.relation_get(unit=unit,
                                            rid=relation_id)
                vault_url = data.get('vault_url')
                role_id = data.get('{}_role_id'.format(hookenv.local_unit()))
                token = data.get('{}_token'.format(hookenv.local_unit()))

                if all([vault_url, role_id, token]):
                    token = json.loads(token)
                    vault_url = json.loads(vault_url)

                    # Tokens may change when secret_id's are being
                    # reissued - if so use token to get new secret_id
                    token_success = False
                    try:
                        secret_id = retrieve_secret_id(
                            url=vault_url,
                            token=token
                        )
                        token_success = True
                    except hvac.exceptions.InvalidRequest:
                        # Try next
                        pass

                    if token_success:
                        db.set('secret-id', secret_id)
                        db.flush()

                        ctxt['vault_url'] = vault_url
                        ctxt['role_id'] = json.loads(role_id)
                        ctxt['secret_id'] = secret_id
                        ctxt['secret_backend'] = self.secret_backend
                        vault_ca = data.get('vault_ca')
                        if vault_ca:
                            ctxt['vault_ca'] = json.loads(vault_ca)

                        self.complete = True
                        break
                    else:
                        if secret_id:
                            ctxt['vault_url'] = vault_url
                            ctxt['role_id'] = json.loads(role_id)
                            ctxt['secret_id'] = secret_id
                            ctxt['secret_backend'] = self.secret_backend
                            vault_ca = data.get('vault_ca')
                            if vault_ca:
                                ctxt['vault_ca'] = json.loads(vault_ca)

            if self.complete:
                break

        if ctxt:
            self.complete = True

        return ctxt


def write_vaultlocker_conf(context, priority=100):
    """Write vaultlocker configuration to disk and install alternative

    :param context: Dict of data from vault-kv relation
    :ptype: context: dict
    :param priority: Priority of alternative configuration
    :ptype: priority: int"""
    charm_vl_path = "/var/lib/charm/{}/vaultlocker.conf".format(
        hookenv.service_name()
    )
    host.mkdir(os.path.dirname(charm_vl_path), perms=0o700)
    templating.render(source='vaultlocker.conf.j2',
                      target=charm_vl_path,
                      context=context, perms=0o600),
    alternatives.install_alternative('vaultlocker.conf',
                                     '/etc/vaultlocker/vaultlocker.conf',
                                     charm_vl_path, priority)


def vault_relation_complete(backend=None):
    """Determine whether vault relation is complete

    :param backend: Name of secrets backend requested
    :ptype backend: string
    :returns: whether the relation to vault is complete
    :rtype: bool"""
    try:
        import hvac
    except ImportError:
        return False
    try:
        vault_kv = VaultKVContext(secret_backend=backend or VAULTLOCKER_BACKEND)
        vault_kv()
        return vault_kv.complete
    except hvac.exceptions.InvalidRequest:
        return False


# TODO: contrib a high level unwrap method to hvac that works
def retrieve_secret_id(url, token):
    """Retrieve a response-wrapped secret_id from Vault

    :param url: URL to Vault Server
    :ptype url: str
    :param token: One shot Token to use
    :ptype token: str
    :returns: secret_id to use for Vault Access
    :rtype: str"""
    import hvac
    try:
        # hvac 0.10.1 changed default adapter to JSONAdapter
        client = hvac.Client(url=url, token=token, adapter=hvac.adapters.Request)
    except AttributeError:
        # hvac < 0.6.2 doesn't have adapter but uses the same response interface
        client = hvac.Client(url=url, token=token)
    else:
        # hvac < 0.9.2 assumes adapter is an instance, so doesn't instantiate
        if not isinstance(client.adapter, hvac.adapters.Request):
            client.adapter = hvac.adapters.Request(base_uri=url, token=token)
    response = client._post('/v1/sys/wrapping/unwrap')
    if response.status_code == 200:
        data = response.json()
        return data['data']['secret_id']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from charmhelpers.core.fstab import Fstab
from nose.tools import (assert_is,
                        assert_is_not,
                        assert_equal)
import unittest
import tempfile
import os

__author__ = 'Jorge Niedbalski R. <jorge.niedbalski@canonical.com>'

DEFAULT_FSTAB_FILE = """/dev/sda /mnt/sda ext3 defaults 0 0
 # This is an indented comment, and the next line is entirely blank.

/dev/sdb /mnt/sdb ext3 defaults 0 0
/dev/sdc	/mnt/sdc	ext3	defaults	0	0
UUID=3af44368-c50b-4768-8e58-aff003cef8be / ext4 errors=remount-ro 0 1
"""

GENERATED_FSTAB_FILE = '\n'.join(
    # Helper will writeback with spaces instead of tabs
    line.replace('\t', ' ') for line in DEFAULT_FSTAB_FILE.splitlines()
    if line.strip() and not line.strip().startswith('#'))


class FstabTest(unittest.TestCase):

    def setUp(self):
        self.tempfile = tempfile.NamedTemporaryFile('w+', delete=False)
        self.tempfile.write(DEFAULT_FSTAB_FILE)
        self.tempfile.close()
        self.fstab = Fstab(path=self.tempfile.name)

    def tearDown(self):
        os.unlink(self.tempfile.name)

    def test_entries(self):
        """Test if entries are correctly read from fstab file"""
        assert_equal(sorted(GENERATED_FSTAB_FILE.splitlines()),
                     sorted(str(entry) for entry in self.fstab.entries))

    def test_get_entry_by_device_attr(self):
        """Test if the get_entry_by_attr method works for device attr"""
        for device in ('sda', 'sdb', 'sdc', ):
            assert_is_not(self.fstab.get_entry_by_attr('device',
                                                       '/dev/%s' % device),
                          None)

    def test_get_entry_by_mountpoint_attr(self):
        """Test if the get_entry_by_attr method works for mountpoint attr"""
        for mnt in ('sda', 'sdb', 'sdc', ):
            assert_is_not(self.fstab.get_entry_by_attr('mountpoint',
                                                       '/mnt/%s' % mnt), None)

    def test_add_entry(self):
        """Test if add_entry works for a new entry"""
        for device in ('sdf', 'sdg', 'sdh'):
            entry = Fstab.Entry('/dev/%s' % device, '/mnt/%s' % device, 'ext3',
                                None)
            assert_is_not(self.fstab.add_entry(entry), None)
            assert_is_not(self.fstab.get_entry_by_attr(
                'device', '/dev/%s' % device), None)

        assert_is(self.fstab.add_entry(entry), False,
                  "Check if adding an existing entry returns false")

    def test_remove_entry(self):
        """Test if remove entry works for already existing entries"""
        for entry in self.fstab.entries:
            assert_is(self.fstab.remove_entry(entry), True)

        assert_equal(len([entry for entry in self.fstab.entries]), 0)
        assert_equal(self.fstab.add_entry(entry), entry)
        assert_equal(len([entry for entry in self.fstab.entries]), 1)

    def test_assert_remove_add_all(self):
        """Test if removing/adding all the entries works"""
        for entry in self.fstab.entries:
            assert_is(self.fstab.remove_entry(entry), True)

        for device in ('sda', 'sdb', 'sdc', ):
            self.fstab.add_entry(
                Fstab.Entry('/dev/%s' % device, '/mnt/%s' % device, 'ext3',
                            None))

        self.fstab.add_entry(Fstab.Entry(
            'UUID=3af44368-c50b-4768-8e58-aff003cef8be',
            '/', 'ext4', 'errors=remount-ro', 0, 1))

        assert_equal(sorted(GENERATED_FSTAB_FILE.splitlines()),
                     sorted(str(entry) for entry in self.fstab.entries))



# Copyright 2017 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import json
from mock import patch
from testtools import TestCase
from tempfile import mkdtemp
from charmhelpers.contrib.storage.linux import bcache

test_stats = {
    'bypassed': '128G\n',
    'cache_bypass_hits': '1132623\n',
    'cache_bypass_misses': '0\n',
    'cache_hit_ratio': '64\n',
    'cache_hits': '12177090\n',
    'cache_miss_collisions': '7091\n',
    'cache_misses': '6717011\n',
    'cache_readaheads': '0\n',
}

tmpdir = 'bcache-stats-test.'
cacheset = 'abcde'
cachedev = 'sdfoo'


class BcacheTestCase(TestCase):
    def setUp(self):
        super(BcacheTestCase, self).setUp()
        self.sysfs = sysfs = mkdtemp(prefix=tmpdir)
        self.addCleanup(shutil.rmtree, sysfs)
        p = patch('charmhelpers.contrib.storage.linux.bcache.SYSFS', new=sysfs)
        p.start()
        self.addCleanup(p.stop)
        self.cacheset = '{}/fs/bcache/{}'.format(sysfs, cacheset)
        os.makedirs(self.cacheset)
        self.devcache = '{}/block/{}/bcache'.format(sysfs, cachedev)
        for n in ['register', 'register_quiet']:
            with open('{}/fs/bcache/{}'.format(sysfs, n), 'w') as f:
                f.write('foo')
        for kind in self.cacheset, self.devcache:
            for sub in bcache.stats_intervals:
                intvaldir = '{}/{}'.format(kind, sub)
                os.makedirs(intvaldir)
                for fn, val in test_stats.items():
                    with open(os.path.join(intvaldir, fn), 'w') as f:
                        f.write(val)

    def test_get_bcache_fs(self):
        bcachedirs = bcache.get_bcache_fs()
        assert len(bcachedirs) == 1
        assert next(iter(bcachedirs)).cachepath.endswith('/fs/bcache/abcde')

    @patch('charmhelpers.contrib.storage.linux.bcache.log', lambda *args, **kwargs: None)
    @patch('charmhelpers.contrib.storage.linux.bcache.os.listdir')
    def test_get_bcache_fs_nobcache(self, mock_listdir):
        mock_listdir.side_effect = OSError(
            '[Errno 2] No such file or directory:...')
        bcachedirs = bcache.get_bcache_fs()
        assert bcachedirs == []

    def test_get_stats_global(self):
        out = bcache.get_stats_action(
            'global', 'hour')
        out = json.loads(out)
        assert len(out.keys()) == 1
        k = next(iter(out.keys()))
        assert k.endswith(cacheset)
        assert out[k]['bypassed'] == '128G'

    def test_get_stats_dev(self):
        out = bcache.get_stats_action(
            cachedev, 'hour')
        out = json.loads(out)
        assert len(out.keys()) == 1
        k = next(iter(out.keys()))
        assert k.endswith('sdfoo/bcache')
        assert out[k]['cache_hit_ratio'] == '64'



import mock
import subprocess
import unittest

from charmhelpers.fetch import ubuntu_apt_pkg as apt_pkg


class Test_apt_pkg_Cache(unittest.TestCase):
    """Borrow PatchHelper methods from ``charms.openstack``."""
    def setUp(self):
        self._patches = {}
        self._patches_start = {}

    def tearDown(self):
        for k, v in self._patches.items():
            v.stop()
            setattr(self, k, None)
        self._patches = None
        self._patches_start = None

    def patch(self, patchee, name=None, **kwargs):
        """Patch a patchable thing.  Uses mock.patch() to do the work.
        Automatically unpatches at the end of the test.

        The mock gets added to the test object (self) using 'name' or the last
        part of the patchee string, after the final dot.

        :param patchee: <string> representing module.object that is to be
            patched.
        :param name: optional <string> name to call the mock.
        :param **kwargs: any other args to pass to mock.patch()
        """
        mocked = mock.patch(patchee, **kwargs)
        if name is None:
            name = patchee.split('.')[-1]
        started = mocked.start()
        self._patches[name] = mocked
        self._patches_start[name] = started
        setattr(self, name, started)

    def patch_object(self, obj, attr, name=None, **kwargs):
        """Patch a patchable thing.  Uses mock.patch.object() to do the work.
        Automatically unpatches at the end of the test.

        The mock gets added to the test object (self) using 'name' or the attr
        passed in the arguments.

        :param obj: an object that needs to have an attribute patched.
        :param attr: <string> that represents the attribute being patched.
        :param name: optional <string> name to call the mock.
        :param **kwargs: any other args to pass to mock.patch()
        """
        mocked = mock.patch.object(obj, attr, **kwargs)
        if name is None:
            name = attr
        started = mocked.start()
        self._patches[name] = mocked
        self._patches_start[name] = started
        setattr(self, name, started)

    def test_apt_cache_show(self):
        self.patch_object(apt_pkg.subprocess, 'check_output')
        apt_cache = apt_pkg.Cache()
        self.check_output.return_value = (
            'Package: dpkg\n'
            'Version: 1.19.0.5ubuntu2.1\n'
            'Bugs: https://bugs.launchpad.net/ubuntu/+filebug\n'
            'Description-en: Debian package management system\n'
            ' Multiline description\n'
            '\n'
            'Package: lsof\n'
            'Architecture: amd64\n'
            'Version: 4.91+dfsg-1ubuntu1\n'
            '\n'
            'N: There is 1 additional record.\n')
        self.assertEquals(
            apt_cache._apt_cache_show(['package']),
            {'dpkg': {
                'package': 'dpkg', 'version': '1.19.0.5ubuntu2.1',
                'bugs': 'https://bugs.launchpad.net/ubuntu/+filebug',
                'description-en': 'Debian package management system\n'
                                  'Multiline description'},
             'lsof': {
                 'package': 'lsof', 'architecture': 'amd64',
                 'version': '4.91+dfsg-1ubuntu1'},
             })
        self.check_output.assert_called_once_with(
            ['apt-cache', 'show', '--no-all-versions', 'package'],
            stderr=subprocess.STDOUT,
            universal_newlines=True)

    def test_dpkg_list(self):
        self.patch_object(apt_pkg.subprocess, 'check_output')
        apt_cache = apt_pkg.Cache()
        self.check_output.return_value = (
            'Desired=Unknown/Install/Remove/Purge/Hold\n'
            '| Status=Not/Inst/Conf-files/Unpacked/halF-conf/Half-inst/'
            'trig-aWait/Trig-pend\n'
            '|/ Err?=(none)/Reinst-required (Status,Err: uppercase=bad)\n'
            '||/ Name           Version            Architecture Description\n'
            '+++-=============================-==================-===========-'
            '=================================\n'
            'ii  dpkg                          1.19.0.5ubuntu2.1  amd64       '
            'Debian package management system\n'
            'rc  linux-image-4.15.0-42-generic 4.15.0-42.45       amd64       '
            'Signed kernel image generic\n'
            'ii  lsof                          4.91+dfsg-1ubuntu1 amd64       '
            'utility to list open files\n')
        expect = {
            'dpkg': {
                'name': 'dpkg',
                'version': '1.19.0.5ubuntu2.1',
                'architecture': 'amd64',
                'description': 'Debian package management system'
            },
            'lsof': {
                'name': 'lsof',
                'version': '4.91+dfsg-1ubuntu1',
                'architecture': 'amd64',
                'description': 'utility to list open files'
            },
        }
        self.assertEquals(
            apt_cache.dpkg_list(['package']), expect)
        self.check_output.side_effect = subprocess.CalledProcessError(
            1, '', output=self.check_output.return_value)
        self.assertEquals(apt_cache.dpkg_list(['package']), expect)
        self.check_output.side_effect = subprocess.CalledProcessError(2, '')
        with self.assertRaises(subprocess.CalledProcessError):
            _ = apt_cache.dpkg_list(['package'])

    def test_version_compare(self):
        self.patch_object(apt_pkg.subprocess, 'check_call')
        self.assertEquals(apt_pkg.version_compare('2', '1'), 1)
        self.check_call.assert_called_once_with(
            ['dpkg', '--compare-versions', '2', 'gt', '1'],
            stderr=subprocess.STDOUT,
            universal_newlines=True)
        self.check_call.side_effect = [
            subprocess.CalledProcessError(1, '', ''),
            None,
            None,
        ]
        self.assertEquals(apt_pkg.version_compare('2', '2'), 0)
        self.check_call.side_effect = [
            subprocess.CalledProcessError(1, '', ''),
            subprocess.CalledProcessError(1, '', ''),
            None,
        ]
        self.assertEquals(apt_pkg.version_compare('1', '2'), -1)
        self.check_call.side_effect = subprocess.CalledProcessError(2, '', '')
        self.assertRaises(subprocess.CalledProcessError,
                          apt_pkg.version_compare, '2', '2')

    def test_apt_cache(self):
        self.patch_object(apt_pkg.subprocess, 'check_output')
        apt_cache = apt_pkg.Cache()
        self.check_output.side_effect = [
            ('Package: dpkg\n'
             'Version: 1.19.0.6ubuntu0\n'
             'Bugs: https://bugs.launchpad.net/ubuntu/+filebug\n'
             'Description-en: Debian package management system\n'
             ' Multiline description\n'
             '\n'
             'Package: lsof\n'
             'Architecture: amd64\n'
             'Version: 4.91+dfsg-1ubuntu1\n'
             '\n'
             'N: There is 1 additional record.\n'),
            ('Desired=Unknown/Install/Remove/Purge/Hold\n'
             '| Status=Not/Inst/Conf-files/Unpacked/halF-conf/Half-inst/'
             'trig-aWait/Trig-pend\n'
             '|/ Err?=(none)/Reinst-required (Status,Err: uppercase=bad)\n'
             '||/ Name           Version            Architecture Description\n'
             '+++-=============================-==================-===========-'
             '=================================\n'
             'ii  dpkg                          1.19.0.5ubuntu2.1  amd64       '
             'Debian package management system\n'
             'rc  linux-image-4.15.0-42-generic 4.15.0-42.45       amd64       '
             'Signed kernel image generic\n'
             'ii  lsof                          4.91+dfsg-1ubuntu1 amd64       '
             'utility to list open files\n'),
        ]
        pkg = apt_cache['dpkg']
        self.assertEquals(pkg.name, 'dpkg')
        self.assertEquals(pkg.current_ver.ver_str, '1.19.0.5ubuntu2.1')
        self.assertEquals(pkg.architecture, 'amd64')
        self.check_output.side_effect = [
            subprocess.CalledProcessError(100, ''),
            subprocess.CalledProcessError(1, ''),
        ]
        with self.assertRaises(KeyError):
            pkg = apt_cache['nonexistent']
        self.check_output.side_effect = [
            ('Package: dpkg\n'
             'Version: 1.19.0.6ubuntu0\n'
             'Bugs: https://bugs.launchpad.net/ubuntu/+filebug\n'
             'Description-en: Debian package management system\n'
             ' Multiline description\n'
             '\n'
             'Package: lsof\n'
             'Architecture: amd64\n'
             'Version: 4.91+dfsg-1ubuntu1\n'
             '\n'
             'N: There is 1 additional record.\n'),
            subprocess.CalledProcessError(42, ''),
        ]
        with self.assertRaises(subprocess.CalledProcessError):
            # System error occurs while making dpkg inquiry
            pkg = apt_cache['dpkg']
        self.check_output.side_effect = [
            subprocess.CalledProcessError(42, ''),
            subprocess.CalledProcessError(1, ''),
        ]
        with self.assertRaises(subprocess.CalledProcessError):
            pkg = apt_cache['system-error-occurs-while-making-apt-inquiry']


class Test_apt_pkg_PkgVersion(unittest.TestCase):

    def test_PkgVersion(self):
        self.assertTrue(
            apt_pkg.PkgVersion('2:20.4.0') <
            apt_pkg.PkgVersion('2:20.4.1'))
        self.assertFalse(
            apt_pkg.PkgVersion('2:20.4.1') <
            apt_pkg.PkgVersion('2:20.4.0'))

        self.assertTrue(
            apt_pkg.PkgVersion('2:20.4.0') <=
            apt_pkg.PkgVersion('2:20.4.1'))
        self.assertTrue(
            apt_pkg.PkgVersion('2:20.4.0') <=
            apt_pkg.PkgVersion('2:20.4.0'))
        self.assertFalse(
            apt_pkg.PkgVersion('2:20.4.1') <=
            apt_pkg.PkgVersion('2:20.4.0'))

        self.assertTrue(
            apt_pkg.PkgVersion('2:20.4.1') >
            apt_pkg.PkgVersion('2:20.4.0'))
        self.assertFalse(
            apt_pkg.PkgVersion('2:20.4.0') >
            apt_pkg.PkgVersion('2:20.4.1'))

        self.assertTrue(
            apt_pkg.PkgVersion('2:20.4.1') >=
            apt_pkg.PkgVersion('2:20.4.0'))
        self.assertTrue(
            apt_pkg.PkgVersion('2:20.4.0') >=
            apt_pkg.PkgVersion('2:20.4.0'))
        self.assertFalse(
            apt_pkg.PkgVersion('2:20.4.0') >=
            apt_pkg.PkgVersion('2:20.4.1'))

        self.assertTrue(
            apt_pkg.PkgVersion('2:20.4.0') ==
            apt_pkg.PkgVersion('2:20.4.0'))
        self.assertFalse(
            apt_pkg.PkgVersion('2:20.4.0') ==
            apt_pkg.PkgVersion('2:20.4.1'))

        self.assertTrue(
            apt_pkg.PkgVersion('2:20.4.0') !=
            apt_pkg.PkgVersion('2:20.4.1'))
        self.assertFalse(
            apt_pkg.PkgVersion('2:20.4.0') !=
            apt_pkg.PkgVersion('2:20.4.0'))

        pkgs = [apt_pkg.PkgVersion('2:20.4.0'),
                apt_pkg.PkgVersion('2:21.4.0'),
                apt_pkg.PkgVersion('2:17.4.0')]
        pkgs.sort()
        self.assertEqual(
            pkgs,
            [apt_pkg.PkgVersion('2:17.4.0'),
             apt_pkg.PkgVersion('2:20.4.0'),
             apt_pkg.PkgVersion('2:21.4.0')])



from mock import patch, call

from testtools import TestCase
from tests.helpers import patch_open, FakeRelation

import charmhelpers.contrib.hahelpers.apache as apache_utils

cert = '''
        -----BEGIN CERTIFICATE-----
MIIDXTCCAkWgAwIBAgIJAMO1fWOu8ntUMA0GCSqGSIb3DQEBCwUAMEUxCzAJBgNV
BAYTAkFVMRMwEQYDVQQIDApTb21lLVN0YXRlMSEwHwYDVQQKDBhJbnRlcm5ldCBX
aWRnaXRzIFB0eSBMdGQwHhcNMTQwNDIyMTUzNDA0WhcNMjQwNDE5MTUzNDA0WjBF
MQswCQYDVQQGEwJBVTETMBEGA1UECAwKU29tZS1TdGF0ZTEhMB8GA1UECgwYSW50
ZXJuZXQgV2lkZ2l0cyBQdHkgTHRkMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIB
CgKCAQEAuk6dmZnMvVxykNidNjbIwXM3ShhMpwCvUmWwpybFAIqhtNTuGJF9Ikp5
kzB+ThQV1onK8O8YarNGQx+MOISEnlJ5npj3Atp33pKGHRn69lHKGVfJvRbN4A90
1hTueYsELzfPV2YWm4c6nQiXRT6Cy0yaw/DE8fBTHzAiE9+/XGPsjn5VPv8H6Wa1
f/d5FblE+RtHP6YpRo9Jh3XAn3iC9fVr8rblS4rk7ev8LfH/yIG2wRVOEPC6lYfu
MEIwPpxKV0c3Z6lqtMOgC5dgzWjrbItnQfB0JaIzSFMMxDhNCJocQRJDQ+0jmj+K
rMGB1QRZlVLZxx0xnv38G0GyfFMv8QIDAQABo1AwTjAdBgNVHQ4EFgQUcxEj7X26
poFDa0lw40aAKIqyNp0wHwYDVR0jBBgwFoAUcxEj7X26poFDa0lw40aAKIqyNp0w
DAYDVR0TBAUwAwEB/zANBgkqhkiG9w0BAQsFAAOCAQEAQe6RUCqTYf0Ns8fKfAEb
QSxZKqCst02oC0F3Gm0opWiUetxZqmAYTAjztmlRFIw7hgF/P95SY1ujGLZmiAlU
poOTjQ/i7MvjkXPVCo92izwXi65qRmJGbjduIirOAYtBmBmm3qS9BmoDlLQMVNYn
bwFImc9ar0h+o3/VH1hry+2vEVikXiKK5uKZI6B7ejNYfAWydq6ilzfKIh75W852
OSbKt3NB/BTZZUdCvK6+B+MoeuzQHDO0/QKBEBfaKFeJki3mdyzFlNbYio1z00rM
E2zl3kh9gkZnMuV1uzHdfKJbtTcNn4hCls5x7T21jn4joADHaVez8FloykBUABu3
qw==
-----END CERTIFICATE-----
'''

IDENTITY_NEW_STYLE_CERTS = {
    'identity-service:0': {
        'keystone/0': {
            'ssl_cert_test-cn': 'keystone_provided_cert',
            'ssl_key_test-cn': 'keystone_provided_key',
        }
    }
}

IDENTITY_OLD_STYLE_CERTS = {
    'identity-service:0': {
        'keystone/0': {
            'ssl_cert': 'keystone_provided_cert',
            'ssl_key': 'keystone_provided_key',
        }
    }
}


class ApacheUtilsTests(TestCase):
    def setUp(self):
        super(ApacheUtilsTests, self).setUp()
        [self._patch(m) for m in [
            'log',
            'config_get',
            'relation_get',
            'relation_ids',
            'relation_list',
            'host',
        ]]

    def _patch(self, method):
        _m = patch.object(apache_utils, method)
        mock = _m.start()
        self.addCleanup(_m.stop)
        setattr(self, method, mock)

    def test_get_cert_from_config(self):
        '''Ensure cert and key from charm config override relation'''
        self.config_get.side_effect = [
            'some_ca_cert',  # config_get('ssl_cert')
            'some_ca_key',  # config_Get('ssl_key')
        ]
        result = apache_utils.get_cert('test-cn')
        self.assertEquals(('some_ca_cert', 'some_ca_key'), result)

    def test_get_ca_cert_from_config(self):
        self.config_get.return_value = "some_ca_cert"
        self.assertEquals('some_ca_cert', apache_utils.get_ca_cert())

    def test_get_cert_from_relation(self):
        self.config_get.return_value = None
        rel = FakeRelation(IDENTITY_NEW_STYLE_CERTS)
        self.relation_ids.side_effect = rel.relation_ids
        self.relation_list.side_effect = rel.relation_units
        self.relation_get.side_effect = rel.get
        result = apache_utils.get_cert('test-cn')
        self.assertEquals(('keystone_provided_cert', 'keystone_provided_key'),
                          result)

    def test_get_cert_from_relation_deprecated(self):
        self.config_get.return_value = None
        rel = FakeRelation(IDENTITY_OLD_STYLE_CERTS)
        self.relation_ids.side_effect = rel.relation_ids
        self.relation_list.side_effect = rel.relation_units
        self.relation_get.side_effect = rel.get
        result = apache_utils.get_cert()
        self.assertEquals(('keystone_provided_cert', 'keystone_provided_key'),
                          result)

    def test_get_ca_cert_from_relation(self):
        self.config_get.return_value = None
        self.relation_ids.side_effect = [['identity-service:0'],
                                         ['identity-credentials:1']]
        self.relation_list.return_value = 'keystone/0'
        self.relation_get.side_effect = [
            'keystone_provided_ca',
        ]
        result = apache_utils.get_ca_cert()
        self.relation_ids.assert_has_calls([call('identity-service'),
                                            call('identity-credentials')])
        self.assertEquals('keystone_provided_ca',
                          result)

    @patch.object(apache_utils.os.path, 'isfile')
    def test_retrieve_ca_cert(self, _isfile):
        _isfile.return_value = True
        with patch_open() as (_open, _file):
            _file.read.return_value = cert
            self.assertEqual(
                apache_utils.retrieve_ca_cert('mycertfile'),
                cert)
            _open.assert_called_once_with('mycertfile', 'rb')

    @patch.object(apache_utils.os.path, 'isfile')
    def test_retrieve_ca_cert_no_file(self, _isfile):
        _isfile.return_value = False
        with patch_open() as (_open, _file):
            self.assertEqual(
                apache_utils.retrieve_ca_cert('mycertfile'),
                None)
            self.assertFalse(_open.called)



from mock import patch

import charmhelpers.contrib.openstack.files.check_deferred_restarts as check_deferred_restarts
from charmhelpers.contrib.openstack.files.check_deferred_restarts import get_deferred_restart_services
import tests.utils


class CheckDeferredRestartsTestCase(tests.utils.BaseTestCase):

    @patch.object(check_deferred_restarts, "get_deferred_events")
    def test_get_deferred_restart_services_no_restarts(self, mock_get_deferred_events):
        mock_get_deferred_events.return_value = [
            {
                "action": "other_action",
                "service": "service1",
            }
        ]

        services = get_deferred_restart_services()
        self.assertListEqual(services, [])

    @patch.object(check_deferred_restarts, "get_deferred_events")
    def test_get_deferred_restart_services_same_application(self, mock_get_deferred_events):
        mock_get_deferred_events.return_value = [
            {
                "action": "restart",
                "service": "service1",
                "policy_requestor_type": "charm",
                "policy_requestor_name": "app1",
            }, {
                "action": "restart",
                "service": "service2",
                "policy_requestor_type": "charm",
                "policy_requestor_name": "app1",
            }
        ]

        services = get_deferred_restart_services()
        self.assertSetEqual(set(services), set(["service1", "service2"]))

        services = get_deferred_restart_services("app1")
        self.assertSetEqual(set(services), set(["service1", "service2"]))

    @patch.object(check_deferred_restarts, "get_deferred_events")
    def test_get_deferred_restart_services_different_applications(self, mock_get_deferred_events):
        mock_get_deferred_events.return_value = [
            {
                "action": "restart",
                "service": "service1",
                "policy_requestor_type": "charm",
                "policy_requestor_name": "app1",
            }, {
                "action": "restart",
                "service": "service2",
                "policy_requestor_type": "charm",
                "policy_requestor_name": "app2",
            }
        ]

        services = get_deferred_restart_services()
        self.assertSetEqual(set(services), set(["service1", "service2"]))

        services = get_deferred_restart_services("app1")
        self.assertSetEqual(set(services), set(["service1"]))

    @patch.object(check_deferred_restarts, "get_deferred_events")
    def test_get_deferred_restart_services_other_requestor(self, mock_get_deferred_events):
        mock_get_deferred_events.return_value = [
            {
                "action": "restart",
                "service": "service1",
                "policy_requestor_type": "charm",
                "policy_requestor_name": "app1",
            }, {
                "action": "restart",
                "service": "service2",
                "policy_requestor_type": "not_charm",
                "policy_requestor_name": "app2",
            }
        ]

        services = get_deferred_restart_services()
        self.assertSetEqual(set(services), set(["service1", "service2"]))

        services = get_deferred_restart_services("app1")
        self.assertSetEqual(set(services), set(["service1", "service2"]))



#!/usr/bin/env python
# coding: utf-8

from unittest import TestCase
from charmhelpers.fetch.python import version

import sys

__author__ = "Jorge Niedbalski <jorge.niedbalski@canonical.com>"


class VersionTestCase(TestCase):

    def setUp(self):
        TestCase.setUp(self)

    def test_current_version(self):
        """
        Check if version.current_version and version.current_version_string
        works correctly
        """
        self.assertEquals(version.current_version(),
                          sys.version_info)
        self.assertEquals(version.current_version_string(),
                          "{0}.{1}.{2}".format(sys.version_info.major,
                                               sys.version_info.minor,
                                               sys.version_info.micro))



# Copyright 2014-2021 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Common python helper functions used for OpenStack charm certificates.

import os
import json
from base64 import b64decode

from charmhelpers.contrib.network.ip import (
    get_hostname,
    resolve_network_cidr,
)
from charmhelpers.core.hookenv import (
    local_unit,
    network_get_primary_address,
    config,
    related_units,
    relation_get,
    relation_ids,
    remote_service_name,
    NoNetworkBinding,
    log,
    WARNING,
    INFO,
)
from charmhelpers.contrib.openstack.ip import (
    resolve_address,
    get_vip_in_network,
    ADDRESS_MAP,
    get_default_api_bindings,
    local_address,
)
from charmhelpers.contrib.network.ip import (
    get_relation_ip,
)

from charmhelpers.core.host import (
    ca_cert_absolute_path,
    install_ca_cert,
    mkdir,
    write_file,
)

from charmhelpers.contrib.hahelpers.apache import (
    CONFIG_CA_CERT_FILE,
)


class CertRequest(object):

    """Create a request for certificates to be generated
    """

    def __init__(self, json_encode=True):
        self.entries = []
        self.hostname_entry = None
        self.json_encode = json_encode

    def add_entry(self, net_type, cn, addresses):
        """Add a request to the batch

        :param net_type: str network space name request is for
        :param cn: str Canonical Name for certificate
        :param addresses: [] List of addresses to be used as SANs
        """
        self.entries.append({
            'cn': cn,
            'addresses': addresses})

    def add_hostname_cn(self):
        """Add a request for the hostname of the machine"""
        ip = local_address(unit_get_fallback='private-address')
        addresses = [ip]
        # If a vip is being used without os-hostname config or
        # network spaces then we need to ensure the local units
        # cert has the appropriate vip in the SAN list
        vip = get_vip_in_network(resolve_network_cidr(ip))
        if vip:
            addresses.append(vip)
        self.hostname_entry = {
            'cn': get_hostname(ip),
            'addresses': addresses}

    def add_hostname_cn_ip(self, addresses):
        """Add an address to the SAN list for the hostname request

        :param addr: [] List of address to be added
        """
        for addr in addresses:
            if addr not in self.hostname_entry['addresses']:
                self.hostname_entry['addresses'].append(addr)

    def get_request(self):
        """Generate request from the batched up entries

        """
        if self.hostname_entry:
            self.entries.append(self.hostname_entry)
        request = {}
        for entry in self.entries:
            sans = sorted(list(set(entry['addresses'])))
            request[entry['cn']] = {'sans': sans}
        if self.json_encode:
            req = {'cert_requests': json.dumps(request, sort_keys=True)}
        else:
            req = {'cert_requests': request}
        req['unit_name'] = local_unit().replace('/', '_')
        return req


def get_certificate_request(json_encode=True, bindings=None):
    """Generate a certificate requests based on the network configuration

    :param json_encode: Encode request in JSON or not. Used for setting
                        directly on a relation.
    :type json_encode: boolean
    :param bindings: List of bindings to check in addition to default api
                     bindings.
    :type bindings: list of strings
    :returns: CertRequest request as dictionary or JSON string.
    :rtype: Union[dict, json]
    """
    if bindings:
        # Add default API bindings to bindings list
        bindings = list(bindings + get_default_api_bindings())
    else:
        # Use default API bindings
        bindings = get_default_api_bindings()
    req = CertRequest(json_encode=json_encode)
    req.add_hostname_cn()
    # Add os-hostname entries
    _sans = get_certificate_sans(bindings=bindings)

    # Handle specific hostnames per binding
    for binding in bindings:
        try:
            hostname_override = config(ADDRESS_MAP[binding]['override'])
        except KeyError:
            hostname_override = None
        try:
            try:
                net_addr = resolve_address(endpoint_type=binding)
            except KeyError:
                net_addr = None
            ip = network_get_primary_address(binding)
            addresses = [net_addr, ip]
            vip = get_vip_in_network(resolve_network_cidr(ip))
            if vip:
                addresses.append(vip)

            # Clear any Nones or duplicates
            addresses = list(set([i for i in addresses if i]))
            # Add hostname certificate request
            if hostname_override:
                req.add_entry(
                    binding,
                    hostname_override,
                    addresses)
                # Remove hostname specific addresses from _sans
                for addr in addresses:
                    try:
                        _sans.remove(addr)
                    except (ValueError, KeyError):
                        pass

        except NoNetworkBinding:
            log("Skipping request for certificate for ip in {} space, no "
                "local address found".format(binding), WARNING)
    # Guarantee all SANs are covered
    # These are network addresses with no corresponding hostname.
    # Add the ips to the hostname cert to allow for this.
    req.add_hostname_cn_ip(_sans)
    return req.get_request()


def get_certificate_sans(bindings=None):
    """Get all possible IP addresses for certificate SANs.

    :param bindings: List of bindings to check in addition to default api
                     bindings.
    :type bindings: list of strings
    :returns: List of binding string names
    :rtype: List[str]
    """
    _sans = [local_address(unit_get_fallback='private-address')]
    if bindings:
        # Add default API bindings to bindings list
        bindings = list(bindings + get_default_api_bindings())
    else:
        # Use default API bindings
        bindings = get_default_api_bindings()

    for binding in bindings:
        # Check for config override
        try:
            net_config = config(ADDRESS_MAP[binding]['config'])
        except KeyError:
            # There is no configuration network for this binding name
            net_config = None
        # Using resolve_address is likely redundant. Keeping it here in
        # case there is an edge case it handles.
        try:
            net_addr = resolve_address(endpoint_type=binding)
        except KeyError:
            net_addr = None
        ip = get_relation_ip(binding, cidr_network=net_config)
        _sans = _sans + [net_addr, ip]
        vip = get_vip_in_network(resolve_network_cidr(ip))
        if vip:
            _sans.append(vip)
    # Clear any Nones and duplicates
    return list(set([i for i in _sans if i]))


def create_ip_cert_links(ssl_dir, custom_hostname_link=None, bindings=None):
    """Create symlinks for SAN records

    :param ssl_dir: str Directory to create symlinks in
    :param custom_hostname_link: str Additional link to be created
    :param bindings: List of bindings to check in addition to default api
                     bindings.
    :type bindings: list of strings
    """

    if bindings:
        # Add default API bindings to bindings list
        bindings = list(bindings + get_default_api_bindings())
    else:
        # Use default API bindings
        bindings = get_default_api_bindings()

    # This includes the hostname cert and any specific bindng certs:
    # admin, internal, public
    req = get_certificate_request(json_encode=False, bindings=bindings)["cert_requests"]
    # Specific certs
    for cert_req in req.keys():
        requested_cert = os.path.join(
            ssl_dir,
            'cert_{}'.format(cert_req))
        requested_key = os.path.join(
            ssl_dir,
            'key_{}'.format(cert_req))
        for addr in req[cert_req]['sans']:
            cert = os.path.join(ssl_dir, 'cert_{}'.format(addr))
            key = os.path.join(ssl_dir, 'key_{}'.format(addr))
            if os.path.isfile(requested_cert) and not os.path.isfile(cert):
                os.symlink(requested_cert, cert)
                os.symlink(requested_key, key)

    # Handle custom hostnames
    hostname = get_hostname(local_address(unit_get_fallback='private-address'))
    hostname_cert = os.path.join(
        ssl_dir,
        'cert_{}'.format(hostname))
    hostname_key = os.path.join(
        ssl_dir,
        'key_{}'.format(hostname))
    if custom_hostname_link:
        custom_cert = os.path.join(
            ssl_dir,
            'cert_{}'.format(custom_hostname_link))
        custom_key = os.path.join(
            ssl_dir,
            'key_{}'.format(custom_hostname_link))
        if os.path.isfile(hostname_cert) and not os.path.isfile(custom_cert):
            os.symlink(hostname_cert, custom_cert)
            os.symlink(hostname_key, custom_key)


def install_certs(ssl_dir, certs, chain=None, user='root', group='root'):
    """Install the certs passed into the ssl dir and append the chain if
       provided.

    :param ssl_dir: str Directory to create symlinks in
    :param certs: {} {'cn': {'cert': 'CERT', 'key': 'KEY'}}
    :param chain: str Chain to be appended to certs
    :param user: (Optional) Owner of certificate files. Defaults to 'root'
    :type user: str
    :param group: (Optional) Group of certificate files. Defaults to 'root'
    :type group: str
    """
    for cn, bundle in certs.items():
        cert_filename = 'cert_{}'.format(cn)
        key_filename = 'key_{}'.format(cn)
        cert_data = bundle['cert']
        if chain:
            # Append chain file so that clients that trust the root CA will
            # trust certs signed by an intermediate in the chain
            cert_data = cert_data + os.linesep + chain
        write_file(
            path=os.path.join(ssl_dir, cert_filename), owner=user, group=group,
            content=cert_data, perms=0o640)
        write_file(
            path=os.path.join(ssl_dir, key_filename), owner=user, group=group,
            content=bundle['key'], perms=0o640)


def get_cert_relation_ca_name(cert_relation_id=None):
    """Determine CA certificate name as provided by relation.

    The filename on disk depends on the name chosen for the application on the
    providing end of the certificates relation.

    :param cert_relation_id: (Optional) Relation id providing the certs
    :type cert_relation_id: str
    :returns: CA certificate filename without path nor extension
    :rtype: str
    """
    if cert_relation_id is None:
        try:
            cert_relation_id = relation_ids('certificates')[0]
        except IndexError:
            return ''
    return '{}_juju_ca_cert'.format(
        remote_service_name(relid=cert_relation_id))


def _manage_ca_certs(ca, cert_relation_id):
    """Manage CA certs.

    :param ca: CA Certificate from certificate relation.
    :type ca: str
    :param cert_relation_id: Relation id providing the certs
    :type cert_relation_id: str
    """
    config_ssl_ca = config('ssl_ca')
    config_cert_file = ca_cert_absolute_path(CONFIG_CA_CERT_FILE)
    if config_ssl_ca:
        log("Installing CA certificate from charm ssl_ca config to {}".format(
            config_cert_file), INFO)
        install_ca_cert(
            b64decode(config_ssl_ca).rstrip(),
            name=CONFIG_CA_CERT_FILE)
    elif os.path.exists(config_cert_file):
        log("Removing CA certificate {}".format(config_cert_file), INFO)
        os.remove(config_cert_file)
    log("Installing CA certificate from certificate relation", INFO)
    install_ca_cert(
        ca.encode(),
        name=get_cert_relation_ca_name(cert_relation_id))


def process_certificates(service_name, relation_id, unit,
                         custom_hostname_link=None, user='root', group='root',
                         bindings=None):
    """Process the certificates supplied down the relation

    :param service_name: str Name of service the certificates are for.
    :param relation_id: str Relation id providing the certs
    :param unit: str Unit providing the certs
    :param custom_hostname_link: str Name of custom link to create
    :param user: (Optional) Owner of certificate files. Defaults to 'root'
    :type user: str
    :param group: (Optional) Group of certificate files. Defaults to 'root'
    :type group: str
    :param bindings: List of bindings to check in addition to default api
                     bindings.
    :type bindings: list of strings
    :returns: True if certificates processed for local unit or False
    :rtype: bool
    """
    if bindings:
        # Add default API bindings to bindings list
        bindings = list(bindings + get_default_api_bindings())
    else:
        # Use default API bindings
        bindings = get_default_api_bindings()

    data = relation_get(rid=relation_id, unit=unit)
    ssl_dir = os.path.join('/etc/apache2/ssl/', service_name)
    mkdir(path=ssl_dir)
    name = local_unit().replace('/', '_')
    certs = data.get('{}.processed_requests'.format(name))
    chain = data.get('chain')
    ca = data.get('ca')
    if certs:
        certs = json.loads(certs)
        _manage_ca_certs(ca, relation_id)
        install_certs(ssl_dir, certs, chain, user=user, group=group)
        create_ip_cert_links(
            ssl_dir,
            custom_hostname_link=custom_hostname_link,
            bindings=bindings)
        return True
    return False


def get_requests_for_local_unit(relation_name=None):
    """Extract any certificates data targeted at this unit down relation_name.

    :param relation_name: str Name of relation to check for data.
    :returns: List of bundles of certificates.
    :rtype: List of dicts
    """
    local_name = local_unit().replace('/', '_')
    raw_certs_key = '{}.processed_requests'.format(local_name)
    relation_name = relation_name or 'certificates'
    bundles = []
    for rid in relation_ids(relation_name):
        for unit in related_units(rid):
            data = relation_get(rid=rid, unit=unit)
            if data.get(raw_certs_key):
                bundles.append({
                    'ca': data['ca'],
                    'chain': data.get('chain'),
                    'certs': json.loads(data[raw_certs_key])})
    return bundles


def get_bundle_for_cn(cn, relation_name=None):
    """Extract certificates for the given cn.

    :param cn: str Canonical Name on certificate.
    :param relation_name: str Relation to check for certificates down.
    :returns: Dictionary of certificate data,
    :rtype: dict.
    """
    entries = get_requests_for_local_unit(relation_name)
    cert_bundle = {}
    for entry in entries:
        for _cn, bundle in entry['certs'].items():
            if _cn == cn:
                cert_bundle = {
                    'cert': bundle['cert'],
                    'key': bundle['key'],
                    'chain': entry['chain'],
                    'ca': entry['ca']}
                break
        if cert_bundle:
            break
    return cert_bundle
